# The MIT License (MIT)
# Copyright (c) 2023 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import abc
import copy
import math
import threading
from typing import Any
from typing import Callable
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pyproj
import xarray as xr

from xcube.util.assertions import assert_given
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from xcube.util.dask import get_block_iterators
from xcube.util.dask import get_chunk_sizes
from xcube.constants import CRS_CRS84
from .helpers import AffineTransformMatrix
from .helpers import Number
from .helpers import _assert_valid_xy_coords
from .helpers import _assert_valid_xy_names
from .helpers import _from_affine
from .helpers import _normalize_int_pair
from .helpers import _normalize_number_pair
from .helpers import _to_affine
from .helpers import scale_xy_res_and_size

# WGS84, axis order: lat, lon
CRS_WGS84 = pyproj.crs.CRS(4326)


# Default tolerance for all operations that
# accept a key-word argument "tolerance":
DEFAULT_TOLERANCE = 1.0e-5


class GridMapping(abc.ABC):
    """
    An abstract base class for grid mappings that define an image grid and
    a transformation from image pixel coordinates to spatial Earth coordinates
    defined in a well-known coordinate reference system (CRS).

    This class cannot be instantiated directly. Use one of its factory methods
    to create instances:

    * :meth:regular()
    * :meth:from_dataset()
    * :meth:from_coords()

    Some instance methods can be used to derive new instances:

    * :meth:derive()
    * :meth:scale()
    * :meth:transform()
    * :meth:to_regular()

    This class is thread-safe.

    """

    def __init__(self,
                 /,
                 size: Union[int, Tuple[int, int]],
                 tile_size: Optional[Union[int, Tuple[int, int]]],
                 xy_bbox: Tuple[Number, Number, Number, Number],
                 xy_res: Union[Number, Tuple[Number, Number]],
                 crs: pyproj.crs.CRS,
                 xy_var_names: Tuple[str, str],
                 xy_dim_names: Tuple[str, str],
                 is_regular: Optional[bool],
                 is_lon_360: Optional[bool],
                 is_j_axis_up: Optional[bool],
                 x_coords: Optional[xr.DataArray] = None,
                 y_coords: Optional[xr.DataArray] = None,
                 xy_coords: Optional[xr.DataArray] = None):

        width, height = _normalize_int_pair(size, name='size')
        assert_true(width > 1 and height > 1,
                    'invalid size')

        tile_width, tile_height = _normalize_int_pair(tile_size,
                                                      default=(width, height))
        assert_true(tile_width > 1 and tile_height > 1,
                    'invalid tile_size')

        assert_given(xy_bbox, name='xy_bbox')
        assert_given(xy_res, name='xy_res')
        _assert_valid_xy_names(xy_var_names, name='xy_var_names')
        _assert_valid_xy_names(xy_dim_names, name='xy_dim_names')
        assert_instance(crs, pyproj.crs.CRS, name='crs')

        if x_coords is not None:
            assert_instance(x_coords, xr.DataArray, name='x_coords')
            assert_true(x_coords.ndim in (1, 2),
                        message=f'x_coords.ndim must be 1 or 2,'
                                f' was {x_coords.ndim}')
        if y_coords is not None:
            assert_instance(y_coords, xr.DataArray, name='y_coords')
            assert_true(y_coords.ndim in (1, 2),
                        message=f'y_coords.ndim must be 1 or 2,'
                                f' was {y_coords.ndim}')
        if xy_coords is not None:
            assert_instance(xy_coords, xr.DataArray, name='xy_coords')
            assert_true(xy_coords.shape == (2, height, width),
                        message=f'xy_coords.shape must be'
                                f' {(2, height, width)},'
                                f' was {xy_coords.shape}')

        x_min, y_min, x_max, y_max = xy_bbox
        x_res, y_res = _normalize_number_pair(xy_res, name='xy_res')
        assert_true(x_res > 0 and y_res > 0,
                    'invalid xy_res')

        self._lock = threading.RLock()

        self._size = width, height
        self._tile_size = tile_width, tile_height
        self._xy_bbox = x_min, y_min, x_max, y_max
        self._xy_res = x_res, y_res
        self._crs = crs
        self._xy_var_names = xy_var_names
        self._xy_dim_names = xy_dim_names
        self._is_regular = is_regular
        self._is_lon_360 = is_lon_360
        self._is_j_axis_up = is_j_axis_up
        self._x_coords = x_coords
        self._y_coords = y_coords
        self._xy_coords = xy_coords

    def derive(self,
               /,
               xy_var_names: Tuple[str, str] = None,
               xy_dim_names: Tuple[str, str] = None,
               tile_size: Union[int, Tuple[int, int]] = None,
               is_j_axis_up: bool = None):
        """
        Derive a new grid mapping from this one with some properties changed.

        :param xy_var_names: The new x-, and y-variable names.
        :param xy_dim_names: The new x-, and y-dimension names.
        :param tile_size: The new tile size
        :param is_j_axis_up: Whether j-axis points up.
        :return: A new, derived grid mapping.
        """
        other = copy.copy(self)
        if xy_var_names is not None:
            _assert_valid_xy_names(xy_var_names, name='xy_var_names')
            other._xy_var_names = xy_var_names
        if xy_dim_names is not None:
            _assert_valid_xy_names(xy_dim_names, name='xy_dim_names')
            other._xy_dim_names = xy_dim_names
        if tile_size is not None:
            tile_width, tile_height = _normalize_int_pair(tile_size,
                                                          name='tile_size')
            assert_true(tile_width > 1 and tile_height > 1,
                        'invalid tile_size')
            tile_size = tile_width, tile_height
            if other.tile_size != tile_size:
                other._tile_size = tile_width, tile_height
                with self._lock:
                    if other._xy_coords is not None:
                        other._xy_coords = other._xy_coords.chunk(
                            other.xy_coords_chunks
                        )
        if is_j_axis_up is not None and is_j_axis_up != other._is_j_axis_up:
            other._is_j_axis_up = is_j_axis_up
            if other._y_coords is not None:
                other._y_coords = other._y_coords[::-1]
            if other._xy_coords is not None:
                other._xy_coords = other._xy_coords[:, ::-1, :]
                other._xy_coords = other._xy_coords.chunk(
                    other.xy_coords_chunks
                )

        return other

    def scale(self,
              xy_scale: Union[Number, Tuple[Number, Number]],
              tile_size: Union[int, Tuple[int, int]] = None) -> 'GridMapping':
        """
        Derive a scaled version of this regular grid mapping.

        Scaling factors lower than one correspond to up-scaling
        (pixels sizes decrease, image size increases).

        Scaling factors larger than one correspond to down-scaling.
        (pixels sizes increase, image size decreases).

        :param xy_scale: The x-, and y-scaling factors.
            May be a single number or tuple.
        :param tile_size: The new tile size
        :return: A new, scaled grid mapping.
        """
        self._assert_regular()
        x_scale, y_scale = _normalize_number_pair(xy_scale)
        new_xy_res, new_size = scale_xy_res_and_size(self.xy_res,
                                                     self.size,
                                                     (x_scale, y_scale))
        if tile_size is not None:
            tile_width, tile_height = _normalize_int_pair(tile_size,
                                                          name='tile_size')
        else:
            tile_width, tile_height = self.tile_size
        tile_width = min(new_size[0], tile_width)
        tile_height = min(new_size[1], tile_height)
        return self.regular(
            new_size,
            (self.x_min, self.y_min),
            new_xy_res,
            self.crs,
            tile_size=(tile_width, tile_height),
            is_j_axis_up=self.is_j_axis_up
        ).derive(
            xy_dim_names=self.xy_dim_names,
            xy_var_names=self.xy_var_names
        )

    @property
    def size(self) -> Tuple[int, int]:
        """Image size (width, height) in pixels."""
        return self._size

    @property
    def width(self) -> int:
        """Image width in pixels."""
        return self.size[0]

    @property
    def height(self) -> int:
        """Image height in pixels."""
        return self.size[1]

    @property
    def tile_size(self) -> Tuple[int, int]:
        """Image tile size (width, height) in pixels."""
        return self._tile_size

    @property
    def is_tiled(self) -> bool:
        """Whether the image is tiled."""
        return self.size != self.tile_size

    @property
    def tile_width(self) -> int:
        """Image tile width in pixels."""
        return self.tile_size[0]

    @property
    def tile_height(self) -> int:
        """Image tile height in pixels."""
        return self.tile_size[1]

    @property
    def x_coords(self):
        """The 1D or 2D x-coordinate array of
        shape (width,) or (height, width).
        """
        return self._get_computed_attribute("_x_coords",
                                            self._new_x_coords)

    @abc.abstractmethod
    def _new_x_coords(self) -> xr.DataArray:
        """Create new 1D or 2D x-coordinate array of
        shape (width,) or (height, width).
        """

    @property
    def y_coords(self):
        """The 1D or 2D y-coordinate array of
        shape (width,) or (height, width).
        """
        return self._get_computed_attribute("_y_coords",
                                            self._new_y_coords)

    @abc.abstractmethod
    def _new_y_coords(self) -> xr.DataArray:
        """Create new 1D or 2D y-coordinate array of
        shape (width,) or (height, width).
        """

    @property
    def xy_coords(self) -> xr.DataArray:
        """
        The x,y coordinates as data array of shape (2, height, width).
        Coordinates are given in units of the CRS.
        """
        xy_coords = self._get_computed_attribute("_xy_coords",
                                                 self._new_xy_coords)
        _assert_valid_xy_coords(xy_coords)
        return xy_coords

    @property
    def xy_coords_chunks(self) -> Tuple[int, int, int]:
        """Get the chunks for the *xy_coords* array."""
        return 2, self.tile_height, self.tile_width

    @abc.abstractmethod
    def _new_xy_coords(self) -> xr.DataArray:
        """Create new coordinate array of shape (2, height, width)."""

    def _get_computed_attribute(self,
                                name: str,
                                computer: Callable[[], Any]) -> Any:
        """Get the value for a computed attribute.
        Utility to be used by this and derived classes.
        """
        value = getattr(self, name)
        if value is not None:
            return value
        with self._lock:
            # Double null check
            value = getattr(self, name)
            if value is not None:
                return value
            value = computer()
            setattr(self, name, value)
            return value

    @property
    def xy_var_names(self) -> Tuple[str, str]:
        """
        The variable names of the x,y coordinates as
        tuple (x_var_name, y_var_name).
        """
        return self._xy_var_names

    @property
    def xy_dim_names(self) -> Tuple[str, str]:
        """
        The dimension names of the x,y coordinates as
        tuple (x_dim_name, y_dim_name).
        """
        return self._xy_dim_names

    @property
    def xy_bbox(self) -> Tuple[float, float, float, float]:
        """The image's bounding box in CRS coordinates."""
        return self._xy_bbox

    @property
    def x_min(self) -> Number:
        """Minimum x-coordinate in CRS units."""
        return self._xy_bbox[0]

    @property
    def y_min(self) -> Number:
        """Minimum y-coordinate in CRS units."""
        return self._xy_bbox[1]

    @property
    def x_max(self) -> Number:
        """Maximum x-coordinate in CRS units."""
        return self._xy_bbox[2]

    @property
    def y_max(self) -> Number:
        """Maximum y-coordinate in CRS units."""
        return self._xy_bbox[3]

    @property
    def xy_res(self) -> Tuple[Number, Number]:
        """Pixel size in x and y direction."""
        return self._xy_res

    @property
    def x_res(self) -> Number:
        """Pixel size in CRS units per pixel in x-direction."""
        return self._xy_res[0]

    @property
    def y_res(self) -> Number:
        """Pixel size in CRS units per pixel in y-direction."""
        return self._xy_res[1]

    @property
    def crs(self) -> pyproj.crs.CRS:
        """The coordinate reference system."""
        return self._crs

    @property
    def spatial_unit_name(self) -> str:
        return self._crs.axis_info[0].unit_name

    @property
    def is_lon_360(self) -> Optional[bool]:
        """
        Check whether *x_max* is greater than 180 degrees.
        Effectively tests whether the range *x_min*, *x_max* crosses
        the anti-meridian at 180 degrees.
        Works only for geographical coordinate reference systems.
        """
        return self._is_lon_360

    @property
    def is_regular(self) -> Optional[bool]:
        """
        Do the x,y coordinates for a regular grid?
        A regular grid has a constant delta in both
        x- and y-directions of the x- and y-coordinates.
        :return None, if this property cannot be determined,
            True or False otherwise.
        """
        return self._is_regular

    @property
    def is_j_axis_up(self) -> Optional[bool]:
        """
        Does the positive image j-axis point up?
        By default, the positive image j-axis points down.
         :return None, if this property cannot be determined,
            True or False otherwise.
        """
        return self._is_j_axis_up

    @property
    def ij_to_xy_transform(self) -> AffineTransformMatrix:
        """
        The affine transformation matrix from image to CRS coordinates.
        Defined only for grid mappings with rectified x,y coordinates.
        """
        self._assert_regular()
        if self.is_j_axis_up:
            return (
                (self.x_res, 0.0, self.x_min),
                (0.0, self.y_res, self.y_min),
            )
        else:
            return (
                (self.x_res, 0.0, self.x_min),
                (0.0, -self.y_res, self.y_max),
            )

    @property
    def xy_to_ij_transform(self) -> AffineTransformMatrix:
        """
        The affine transformation matrix from CRS to image coordinates.
        Defined only for grid mappings with rectified x,y coordinates.
        """
        self._assert_regular()
        return _from_affine(~_to_affine(self.ij_to_xy_transform))

    def ij_transform_to(self, other: 'GridMapping') -> AffineTransformMatrix:
        """
        Get the affine transformation matrix that transforms
        image coordinates of *other* into image coordinates
        of this image geometry.

        Defined only for grid mappings with rectified x,y coordinates.

        :param other: The other image geometry
        :return: Affine transformation matrix
        """
        self._assert_regular()
        self.assert_regular(other, name='other')
        a = _to_affine(self.ij_to_xy_transform)
        b = _to_affine(other.xy_to_ij_transform)
        return _from_affine(b * a)

    def ij_transform_from(self, other: 'GridMapping') -> AffineTransformMatrix:
        """
        Get the affine transformation matrix that transforms
        image coordinates of this image geometry to image
        coordinates of *other*.

        Defined only for grid mappings with rectified x,y coordinates.

        :param other: The other image geometry
        :return: Affine transformation matrix
        """
        self._assert_regular()
        self.assert_regular(other, name='other')
        a = _to_affine(self.ij_transform_to(other))
        return _from_affine(~a)

    @property
    def ij_bbox(self) -> Tuple[int, int, int, int]:
        """The image's bounding box in pixel coordinates."""
        return 0, 0, self.width, self.height

    @property
    def ij_bboxes(self) -> np.ndarray:
        """The image tiles' bounding boxes in image pixel coordinates."""
        chunk_sizes = get_chunk_sizes((self.height, self.width),
                                      (self.tile_height, self.tile_width))
        _, _, block_slices = get_block_iterators(chunk_sizes)
        block_slices = tuple(block_slices)
        n = len(block_slices)
        ij_bboxes = np.ndarray((n, 4), dtype=np.int64)
        for i in range(n):
            y_slice, x_slice = block_slices[i]
            ij_bboxes[i, 0] = x_slice.start
            ij_bboxes[i, 1] = y_slice.start
            ij_bboxes[i, 2] = x_slice.stop
            ij_bboxes[i, 3] = y_slice.stop
        return ij_bboxes

    @property
    def xy_bboxes(self) -> np.ndarray:
        """The image tiles' bounding boxes in CRS coordinates."""
        if self.is_j_axis_up:
            xy_offset = np.array([self.x_min, self.y_min,
                                  self.x_min, self.y_min])
            xy_scale = np.array([self.x_res, self.y_res,
                                 self.x_res, self.y_res])
            xy_bboxes = xy_offset + xy_scale * self.ij_bboxes
        else:
            xy_offset = np.array([self.x_min, self.y_max,
                                  self.x_min, self.y_max])
            xy_scale = np.array([self.x_res, -self.y_res,
                                 self.x_res, -self.y_res])
            xy_bboxes = xy_offset + xy_scale * self.ij_bboxes
            xy_bboxes[:, [1, 3]] = xy_bboxes[:, [3, 1]]
        return xy_bboxes

    def ij_bbox_from_xy_bbox(self,
                             xy_bbox: Tuple[float, float, float, float],
                             xy_border: float = 0.0,
                             ij_border: int = 0) \
            -> Tuple[int, int, int, int]:
        """
        Compute bounding box in i,j pixel coordinates given a
        bounding box *xy_bbox* in x,y coordinates.

        :param xy_bbox: Box (x_min, y_min, x_max, y_max) given in
            the same CS as x and y.
        :param xy_border: If non-zero, grows the bounding box *xy_bbox*
            before using it for comparisons. Defaults to 0.
        :param ij_border: If non-zero, grows the returned i,j
            bounding box and clips it to size. Defaults to 0.
        :return: Bounding box in (i_min, j_min, i_max, j_max)
            in pixel coordinates. Returns ``(-1, -1, -1, -1)``
            if *xy_bbox* isn't intersecting any of the x,y coordinates.
        """
        xy_bboxes = np.array([xy_bbox], dtype=np.float64)
        ij_bboxes = np.full_like(xy_bboxes, -1, dtype=np.int64)
        self.ij_bboxes_from_xy_bboxes(xy_bboxes,
                                      xy_border=xy_border,
                                      ij_border=ij_border,
                                      ij_bboxes=ij_bboxes)
        # noinspection PyTypeChecker
        return tuple(map(int, ij_bboxes[0]))

    def ij_bboxes_from_xy_bboxes(self,
                                 xy_bboxes: np.ndarray,
                                 xy_border: float = 0.0,
                                 ij_border: int = 0,
                                 ij_bboxes: np.ndarray = None) \
            -> np.ndarray:
        """
        Compute bounding boxes in pixel coordinates given bounding boxes
        *xy_bboxes* [[x_min, y_min, x_max, y_max], ...] in x,y coordinates.

        The returned array in i,j pixel coordinates
        has the same shape as *xy_bboxes*. The value ranges in the
        returned array [[i_min, j_min, i_max, j_max], ..]] are:

        * i_min from 0 to width-1, i_max from 1 to width;
        * j_min from 0 to height-1, j_max from 1 to height;

        so the i,j pixel coordinates can be used as array index slices.

        :param xy_bboxes: Numpy array of x,y bounding boxes
            [[x_min, y_min, x_max, y_max], ...] given in the same
            CS as x and y.
        :param xy_border: If non-zero, grows the bounding box *xy_bbox*
            before using it for comparisons. Defaults to 0.
        :param ij_border: If non-zero, grows the returned i,j
            bounding box and clips it to size. Defaults to 0.
        :param ij_bboxes: Numpy array of pixel i,j bounding boxes
            [[x_min, y_min, x_max, y_max], ...]. If given, must have
            same shape as *xy_bboxes*.
        :return: Bounding boxes in [[i_min, j_min, i_max, j_max], ..]]
            in pixel coordinates.
        """
        from .bboxes import compute_ij_bboxes
        if ij_bboxes is None:
            ij_bboxes = np.full_like(xy_bboxes, -1, dtype=np.int64)
        else:
            ij_bboxes[:, :] = -1
        xy_coords = self.xy_coords.values
        compute_ij_bboxes(xy_coords[0],
                          xy_coords[1],
                          xy_bboxes,
                          xy_border,
                          ij_border,
                          ij_bboxes)
        return ij_bboxes

    def to_dataset_attrs(self) -> Mapping[str, Any]:
        """
        Get spatial dataset attributes as recommended by
        https://wiki.esipfed.org/Attribute_Convention_for_Data_Discovery_1-3#Recommended

        :return: dictionary with dataset coordinate attributes.
        """

        x1, y1, x2, y2 = self.xy_bbox

        if self.crs.is_geographic:
            lon_min, lat_min, lon_max, lat_max = self.xy_bbox
            lon_res, lat_res = self.xy_res
        else:
            x_res, y_res = self.xy_res
            # center position
            xm1 = (x1 + x2) / 2
            ym1 = (y1 + y2) / 2
            # center position + delta
            xm2 = xm1 + x_res
            ym2 = ym1 + y_res
            transformer = pyproj.Transformer.from_crs(crs_from=self.crs,
                                                      crs_to=CRS_CRS84)
            xx, yy = transformer.transform((x1, x2, xm1, xm2),
                                           (y1, y2, ym1, ym2))
            lon_min, lon_max, lon_m1, lon_m2 = xx
            lat_min, lat_max, lat_m1, lat_m2 = yy
            # Estimate resolution (note, this may be VERY wrong)
            lon_res = abs(lon_m2 - lon_m1)
            lat_res = abs(lat_m2 - lat_m1)

        geospatial_bounds_crs = 'CRS84'
        geospatial_bounds = (f'POLYGON(('
                             f'{lon_min} {lat_min}, '
                             f'{lon_min} {lat_max}, '
                             f'{lon_max} {lat_max}, '
                             f'{lon_max} {lat_min}, '
                             f'{lon_min} {lat_min}'
                             f'))')

        return dict(
            geospatial_lon_units='degrees_east',
            geospatial_lon_min=lon_min,
            geospatial_lon_max=lon_max,
            geospatial_lon_resolution=lon_res,
            geospatial_lat_units='degrees_north',
            geospatial_lat_min=lat_min,
            geospatial_lat_max=lat_max,
            geospatial_lat_resolution=lat_res,
            geospatial_bounds_crs=geospatial_bounds_crs,
            geospatial_bounds=geospatial_bounds,
        )

    def to_coords(self,
                  xy_var_names: Tuple[str, str] = None,
                  xy_dim_names: Tuple[str, str] = None,
                  exclude_bounds: bool = False,
                  reuse_coords: bool = False,
                  ) \
            -> Mapping[str, xr.DataArray]:
        """
        Get CF-compliant axis coordinate variables and cell boundary
        coordinate variables.

        Defined only for grid mappings with regular x,y coordinates.

        :param xy_var_names: Optional coordinate variable
            names (x_var_name, y_var_name).
        :param xy_dim_names: Optional coordinate dimensions
            names (x_dim_name, y_dim_name).
        :param exclude_bounds: If True, do not create bounds
            coordinates. Defaults to False.
        :param reuse_coords: Whether to either reuse target
            coordinate arrays from target_gm or to compute
            new ones.
        :return: dictionary with coordinate variables
        """
        self._assert_regular()
        from .coords import grid_mapping_to_coords
        return grid_mapping_to_coords(self,
                                      xy_var_names=xy_var_names,
                                      xy_dim_names=xy_dim_names,
                                      exclude_bounds=exclude_bounds,
                                      reuse_coords=reuse_coords)

    def transform(self,
                  crs: Union[str, pyproj.crs.CRS],
                  *,
                  tile_size: Union[int, Tuple[int, int]] = None,
                  xy_var_names: Tuple[str, str] = None,
                  tolerance: float = DEFAULT_TOLERANCE) \
            -> 'GridMapping':
        """
        Transform this grid mapping so it uses the given
        spatial coordinate reference system into another *crs*.

        :param crs: The new spatial coordinate reference system.
        :param tile_size: Optional new tile size.
        :param xy_var_names: Optional new coordinate names.
        :param tolerance: Absolute tolerance used when comparing
            coordinates with each other. Must be in the units
            of the *crs* and must be greater zero.
        :return: A new grid mapping that uses *crs*.
        """
        from .transform import transform_grid_mapping
        return transform_grid_mapping(self,
                                      crs,
                                      tile_size=tile_size,
                                      xy_var_names=xy_var_names,
                                      tolerance=tolerance)

    @classmethod
    def regular(cls,
                size: Union[int, Tuple[int, int]],
                xy_min: Tuple[float, float],
                xy_res: Union[float, Tuple[float, float]],
                crs: Union[str, pyproj.crs.CRS],
                *,
                tile_size: Union[int, Tuple[int, int]] = None,
                is_j_axis_up: bool = False) -> 'GridMapping':
        """
        Create a new regular grid mapping.

        :param size: Size in pixels.
        :param xy_min: Minimum x- and y-coordinates.
        :param xy_res: Resolution in x- and y-directions.
        :param crs: Spatial coordinate reference system.
        :param tile_size: Optional tile size.
        :param is_j_axis_up: Whether positive j-axis points up.
            Defaults to false.
        :return: A new regular grid mapping.
        """
        from .regular import new_regular_grid_mapping
        return new_regular_grid_mapping(size=size,
                                        xy_min=xy_min,
                                        xy_res=xy_res,
                                        crs=crs,
                                        tile_size=tile_size,
                                        is_j_axis_up=is_j_axis_up)

    def to_regular(self,
                   tile_size: Union[int, Tuple[int, int]] = None,
                   is_j_axis_up: bool = False) -> 'GridMapping':
        """
        Transform this grid mapping into one that is regular.

        :param tile_size: Optional tile size.
        :param is_j_axis_up: Whether positive j-axis points up.
            Defaults to false.
        :return: A new regular grid mapping or this grid mapping,
            if it is already regular.
        """
        from .regular import to_regular_grid_mapping
        return to_regular_grid_mapping(self,
                                       tile_size=tile_size,
                                       is_j_axis_up=is_j_axis_up)

    @classmethod
    def from_dataset(cls,
                     dataset: xr.Dataset,
                     *,
                     crs: Union[str, pyproj.crs.CRS] = None,
                     tile_size: Union[int, Tuple[int, int]] = None,
                     prefer_is_regular: bool = True,
                     prefer_crs: Union[str, pyproj.crs.CRS] = None,
                     emit_warnings: bool = False,
                     tolerance: float = DEFAULT_TOLERANCE) -> 'GridMapping':
        """
        Create a grid mapping for the given *dataset*.

        :param dataset: The dataset.
        :param crs: Optional spatial coordinate reference system.
        :param tile_size: Optional tile size
        :param prefer_is_regular: Whether to prefer a regular
            grid mapping if multiple found. Default is True.
        :param prefer_crs: The preferred CRS of a grid mapping
            if multiple found.
        :param emit_warnings: Whether to emit warning for
            non-CF compliant datasets.
        :param tolerance: Absolute tolerance used when comparing
            coordinates with each other. Must be in the units
            of the *crs* and must be greater zero.
        :return: a new grid mapping instance.
        """
        from .dataset import new_grid_mapping_from_dataset
        return new_grid_mapping_from_dataset(
            dataset=dataset,
            crs=crs,
            tile_size=tile_size,
            prefer_is_regular=prefer_is_regular,
            prefer_crs=prefer_crs,
            emit_warnings=emit_warnings,
            tolerance=tolerance
        )

    @classmethod
    def from_coords(cls,
                    x_coords: xr.DataArray,
                    y_coords: xr.DataArray,
                    crs: Union[str, pyproj.crs.CRS],
                    *,
                    tile_size: Union[int, Tuple[int, int]] = None,
                    tolerance: float = DEFAULT_TOLERANCE) \
            -> 'GridMapping':
        """
        Create a grid mapping from given x- and y-coordinates
        *x_coords*, *y_coords* and spatial coordinate reference
        system *crs*.

        :param x_coords: The x-coordinates.
        :param y_coords: The y-coordinates.
        :param crs: The spatial coordinate reference system.
        :param tile_size: Optional tile size.
        :param tolerance: Absolute tolerance used when comparing
            coordinates with each other. Must be in the units
            of the *crs* and must be greater zero.
        :return: A new grid mapping.
        """
        from .coords import new_grid_mapping_from_coords
        return new_grid_mapping_from_coords(x_coords=x_coords,
                                            y_coords=y_coords,
                                            crs=crs,
                                            tile_size=tile_size,
                                            tolerance=tolerance)

    def is_close(self,
                 other: 'GridMapping',
                 tolerance: float = DEFAULT_TOLERANCE) -> bool:
        """
        Tests whether this grid mapping is close to *other*.

        :param other: The other grid mapping.
        :param tolerance: Absolute tolerance used when comparing
            coordinates with each other. Must be in the units
            of the *crs* and must be greater zero.
        :return: True, if so, False otherwise.
        """
        if self is other:
            return True
        if self.is_j_axis_up == other.is_j_axis_up \
                and self.is_lon_360 == other.is_lon_360 \
                and self.is_regular == other.is_regular \
                and self.size == other.size \
                and self.tile_size == other.tile_size \
                and self.crs == other.crs:
            sxr, syr = self.xy_res
            oxr, oyr = other.xy_res
            if math.isclose(sxr, oxr, abs_tol=tolerance) \
                    and math.isclose(syr, oyr, abs_tol=tolerance):
                sx1, sy1, sx2, sy2 = self.xy_bbox
                ox1, oy1, ox2, oy2 = other.xy_bbox
                return math.isclose(sx1, ox1, abs_tol=tolerance) \
                    and math.isclose(sy1, oy1, abs_tol=tolerance) \
                    and math.isclose(sx2, ox2, abs_tol=tolerance) \
                    and math.isclose(sy2, oy2, abs_tol=tolerance)
        return False

    @classmethod
    def assert_regular(cls, value: Any, name: str = None):
        assert_instance(value, GridMapping, name=name)
        if not value.is_regular:
            raise ValueError(f'{name or "value"} must'
                             f' be a regular grid mapping')

    def _assert_regular(self):
        if not self.is_regular:
            raise NotImplementedError('Operation not implemented'
                                      ' for non-regular grid mappings')

    def _repr_markdown_(self) -> str:
        """Generate an IPython Notebook Markdown representation."""
        is_regular = self.is_regular \
            if self.is_regular is not None else "_unknown_"
        is_j_axis_up = self.is_j_axis_up \
            if self.is_j_axis_up is not None else "_unknown_"
        is_lon_360 = self.is_lon_360 \
            if self.is_lon_360 is not None else "_unknown_"
        xy_res = repr(self.xy_res) \
                 + ('' if self.is_regular else '  _estimated_')
        return '\n'.join([
            f'class: **{self.__class__.__name__}**',
            f'* is_regular: {is_regular}',
            f'* is_j_axis_up: {is_j_axis_up}',
            f'* is_lon_360: {is_lon_360}',
            f'* crs: {self.crs}',
            f'* xy_res: {xy_res}',
            f'* xy_bbox: {self.xy_bbox}',
            f'* ij_bbox: {self.ij_bbox}',
            f'* xy_dim_names: {self.xy_dim_names}',
            f'* xy_var_names: {self.xy_var_names}',
            f'* size: {self.size}',
            f'* tile_size: {self.tile_size}',
        ])



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import inspect
from typing import Callable, Any, Dict

from xcube.server.api import ApiError

from .context import ComputeContext
from .op.info import OpInfo


def get_compute_operations(ctx: ComputeContext):
    ops = ctx.op_registry.ops
    return {
        "operations": [encode_op(op_id, f) for op_id, f in ops.items()]
    }


def get_compute_operation(ctx: ComputeContext, op_id: str):
    op = ctx.op_registry.get_op(op_id)
    if op is None:
        raise ApiError.NotFound(f'operation {op_id!r} not found')
    return encode_op(op_id, op)


def encode_op(op_id: str, op: Callable) -> Dict[str, Any]:
    op_info = OpInfo.get_op_info(op)
    op_json = {
        "operationId": op_id,
        "parametersSchema": op_info.params_schema
    }
    doc = inspect.getdoc(op)
    if doc:
        op_json.update(description=doc)
    return op_json



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import math
from typing import Optional, Tuple, Sequence, List, Union

import numpy as np
import pyproj

from xcube.util.assertions import assert_given
from xcube.util.assertions import assert_instance
from xcube.util.types import Pair
from xcube.util.types import ScalarOrPair
from xcube.util.types import normalize_scalar_or_pair

WEB_MERCATOR_CRS_NAME = 'EPSG:3857'
WEB_MERCATOR_CRS_ALIASES = (
    'EPSG:3857',
    'urn:ogc:def:crs:OGC:1.3:EPSG::3857',
)
GEOGRAPHIC_CRS_NAME = 'OGC:CRS84'
GEOGRAPHIC_CRS_ALIASES = (
    'EPSG:4326',
    'urn:ogc:def:crs:OGC:1.3:EPSG::4326',
    'OGC:CRS84',
    'CRS84',
    'urn:ogc:def:crs:OGC:1.3:CRS84',
    'WGS84'
)
DEFAULT_CRS_NAME = GEOGRAPHIC_CRS_NAME
DEFAULT_TILE_SIZE = 256
EARTH_EQUATORIAL_RADIUS_WGS84 = 6378137.
EARTH_CIRCUMFERENCE_WGS84 = 2 * math.pi * EARTH_EQUATORIAL_RADIUS_WGS84


class TilingScheme:
    """
    A scheme for subdividing the world into a tiled, multi-resolution
    image pyramid.

    Two standard tiling schemes are pre-defined:

    1. TilingScheme.WEB_MERCATOR
    2. TilingScheme.GEOGRAPHIC

    :param num_level_zero_tiles: The number of tiles
        in x and y direction at lowest resolution level (zero).
    :param crs_name: Name of the spatial coordinate reference system.
    :param map_height: The height of the world map in units of the
        spatial coordinate reference system.
    :param tile_size: The tile size to be used for
        both x and y directions. Defaults to 256.
    :param min_level: Optional minimum resolution level.
    :param max_level: Optional maximum resolution level.
    """

    # The web mercator tiling schema with 1 x 1 level zero tiles
    WEB_MERCATOR: 'TilingScheme'
    # The geographic tiling schema with 2 x 1 level zero tiles
    GEOGRAPHIC: 'TilingScheme'

    def __init__(self,
                 num_level_zero_tiles: Tuple[int, int],
                 crs_name: str,
                 map_height: float,
                 tile_size: ScalarOrPair[int] = DEFAULT_TILE_SIZE,
                 min_level: Optional[int] = None,
                 max_level: Optional[int] = None):
        tile_height, tile_width = normalize_scalar_or_pair(tile_size)
        self._num_level_zero_tiles = num_level_zero_tiles
        self._crs_name = crs_name
        self._crs = None
        self._map_height = map_height
        self._tile_width = tile_width
        self._tile_height = tile_height
        self._min_level = min_level
        self._max_level = max_level

    @property
    def num_level_zero_tiles(self) -> Pair[int]:
        """The number of level zero tiles in x and y directions."""
        return self._num_level_zero_tiles

    @property
    def level_zero_resolution(self) -> float:
        """The (vertical) resolution at level zero in map units."""
        return self._map_height / self._tile_height

    @property
    def crs_name(self) -> str:
        """The name of the spatial coordinate reference system."""
        return self._crs_name

    @property
    def crs(self) -> pyproj.CRS:
        """The spatial coordinate reference system."""
        if self._crs is None:
            self._crs = pyproj.CRS.from_string(self.crs_name)
        return self._crs

    @property
    def map_unit_name(self) -> str:
        """The name of the map's spatial units."""
        return self.crs.axis_info[0].unit_name

    @property
    def map_width(self) -> float:
        """
        The height of the map in units
        of the map's spatial coordinate reference system.
        """
        num_tiles_x0, num_tiles_y0 = self.num_level_zero_tiles
        return self.map_height * num_tiles_x0 / num_tiles_y0

    @property
    def map_height(self) -> float:
        """
        The height of the map in units
        of the map's spatial coordinate reference system.
        """
        return self._map_height

    @property
    def map_extent(self) -> Tuple[float, float, float, float]:
        """
        The extent of the map in units
        of the map's spatial coordinate reference system.
        """
        map_width_05 = self.map_width / 2
        map_height_05 = self.map_height / 2
        return -map_width_05, -map_height_05, map_width_05, map_height_05

    @property
    def map_origin(self) -> Tuple[float, float]:
        """
        The origin of the map (upper, left pixel of the upper left tile)
        in units of the map's spatial coordinate reference system.
        """
        map_extent = self.map_extent
        return map_extent[0], map_extent[3]

    @property
    def tile_size(self) -> Pair[int]:
        """The tile size in pixels."""
        return self._tile_width, self._tile_height

    @property
    def tile_width(self) -> int:
        """The tile width in pixels."""
        return self._tile_width

    @property
    def tile_height(self) -> int:
        """The tile height in pixels."""
        return self._tile_height

    @property
    def min_level(self) -> Optional[int]:
        """The minimum level of detail."""
        return self._min_level

    @property
    def max_level(self) -> Optional[int]:
        """The maximum level of detail."""
        return self._max_level

    @property
    def num_levels(self) -> Optional[int]:
        """The number of detail levels."""
        return self.max_level + 1 if self.max_level is not None else None

    @classmethod
    def for_crs(cls, crs: Union[pyproj.CRS, str], **kwargs) -> 'TilingScheme':
        """
        Get a new tiling scheme for the named coordinate reference system.

        :param crs: The spatial coordinate reference system.
        :param kwargs: subset of constructor arguments
        :return: a tiling scheme
        """
        assert_instance(crs, (pyproj.CRS, str), name='crs')
        crs_name = crs.srs if isinstance(crs, pyproj.CRS) else crs
        if crs_name in WEB_MERCATOR_CRS_ALIASES:
            tiling_scheme = cls.WEB_MERCATOR
        elif crs_name in GEOGRAPHIC_CRS_ALIASES:
            tiling_scheme = cls.GEOGRAPHIC
        else:
            raise ValueError(f'unsupported spatial CRS {crs_name!r}')
        return tiling_scheme.derive(**kwargs) if kwargs else tiling_scheme

    def derive(self, **kwargs):
        """
        Derive a new tiling scheme using a subset of constructor
        arguments *kwargs*.

        :param kwargs: subset of constructor arguments
        :return: a new tiling scheme
        """
        args = self.to_dict()
        args.update({k: v for k, v in kwargs.items() if v is not None})
        return TilingScheme(**args)

    def to_dict(self):
        """
        Return a JSON-serializable dictionary
        representation of this tiling scheme.
        """
        d = dict(num_level_zero_tiles=self.num_level_zero_tiles,
                 crs_name=self.crs_name,
                 map_height=self.map_height,
                 tile_size=self.tile_size,
                 min_level=self.min_level,
                 max_level=self.max_level)
        return {k: v for k, v in d.items() if v is not None}

    def get_resolutions(self,
                        min_level: Optional[int] = None,
                        max_level: Optional[int] = None,
                        unit_name: Optional[str] = None) -> List[float]:
        """
        Get spatial resolutions for this tiling scheme.

        :param min_level: Optional minimum level. Defaults to self.min_level.
        :param max_level: Optional maximum level. Defaults to self.max_level.
        :param unit_name: The spatial unit for the returned resolutions.
        :return: List of spatial resolutions.
        """
        unit_factor = get_unit_factor(self.map_unit_name,
                                      unit_name or self.map_unit_name)
        res_l0 = unit_factor * self.level_zero_resolution

        min_level = min_level if min_level is not None else self.min_level
        if min_level is None:
            min_level = 0
        max_level = max_level if max_level is not None else self.max_level
        if max_level is None:
            raise ValueError('max_value must be given')

        return [res_l0 / (1 << level)
                for level in range(min_level, max_level + 1)]

    def get_levels_for_resolutions(
            self,
            resolutions: Sequence[float],
            unit_name: str
    ) -> Tuple[int, int]:
        """
        Get the minimum and maximum level indices into this tiling scheme
        for the given spatial *resolutions*.

        The resolutions are typically those of a multi-resolution dataset,
        where the first entry represents level zero, the highest resolution,
        hence the smallest resolution value. Subsequent resolution values
        are monotonically increasing.

        :param resolutions: A sequence of spatial resolutions.
        :param unit_name: The spatial units for *resolutions*.
        :return: the range of levels in this tiling scheme.
        """
        assert_given(resolutions, name='resolutions')
        assert_instance(unit_name, str, name='unit_name')
        if not isinstance(resolutions, np.ndarray):
            resolutions = np.array(resolutions)

        f_to_map = get_unit_factor(unit_name, self.map_unit_name)
        map_resolutions = f_to_map * resolutions
        map_levels = np.ceil(np.log2(
            self.level_zero_resolution / map_resolutions
        ))
        min_level = max(int(np.min(map_levels)), 0)
        max_level = max(int(np.max(map_levels)), min_level)

        return min_level, max_level

    def get_resolutions_level(
            self,
            level: int,
            resolutions: Sequence[float],
            unit_name: str
    ) -> int:
        """
        Get the level in the sequence of spatial *resolutions*
        for given *map_level*.

        The resolutions are typically those of a multi-resolution dataset,
        where the first entry represents level zero, the highest resolution,
        hence the smallest resolution value. Subsequent resolution values
        are monotonically increasing.

        :param level: A level within this tiling scheme.
        :param resolutions: A sequence of spatial resolutions.
            Values must be monotonically increasing. First entry
            is the highest resolution at level zero.
        :param unit_name: The spatial units for *resolutions*.
        :return: The multi-resolution level.
        """
        assert_given(resolutions, name='resolutions')
        assert_instance(unit_name, str, name='unit_name')

        f_from_map = get_unit_factor(self.map_unit_name, unit_name)
        # Tile pixel size in dataset units for map tile at level 0
        ds_pix_size_l0 = f_from_map * self.level_zero_resolution
        # Tile pixel size in dataset units for map tile at level
        ds_pix_size = ds_pix_size_l0 / (1 << level)

        num_ds_levels = len(resolutions)

        ds_pix_size_min = resolutions[0]
        if ds_pix_size <= ds_pix_size_min:
            return 0

        ds_pix_size_max = resolutions[-1]
        if ds_pix_size >= ds_pix_size_max:
            return num_ds_levels - 1

        for ds_level in range(num_ds_levels - 1):
            ds_pix_size_1 = resolutions[ds_level]
            ds_pix_size_2 = resolutions[ds_level + 1]
            if ds_pix_size_1 <= ds_pix_size <= ds_pix_size_2:
                r = (ds_pix_size - ds_pix_size_1) \
                    / (ds_pix_size_2 - ds_pix_size_1)
                if r < 0.5:
                    return ds_level
                else:
                    return ds_level + 1

        raise RuntimeError('should not come here')

    def get_tile_extent(self,
                        tile_x: int,
                        tile_y: int,
                        tile_z: int) \
            -> Optional[Tuple[float, float, float, float]]:
        """
        Get the extent in units of the CRS for the given tile coordinates.

        :param tile_x: The tile's column index
        :param tile_y: The tile's row index
        :param tile_z: The tile's level index
        :return: The tile's extent
        """
        if tile_z < 0:
            return None

        zoom_factor = 1 << tile_z

        num_tiles_x0, num_tiles_y0 = self.num_level_zero_tiles

        num_tiles_x = num_tiles_x0 * zoom_factor
        if tile_x < 0 or tile_x >= num_tiles_x:
            return None

        num_tiles_y = num_tiles_y0 * zoom_factor
        if tile_y < 0 or tile_y >= num_tiles_y:
            return None

        map_width = self.map_width
        map_height = self.map_height

        map_x0 = -map_width / 2
        map_y0 = map_height / 2

        map_tile_width = map_width / zoom_factor / num_tiles_x0
        map_tile_height = map_height / zoom_factor / num_tiles_y0

        x1 = map_x0 + tile_x * map_tile_width
        y1 = map_y0 - (tile_y + 1) * map_tile_height

        x2 = map_x0 + (tile_x + 1) * map_tile_width
        y2 = map_y0 - tile_y * map_tile_height

        return x1, y1, x2, y2


TilingScheme.WEB_MERCATOR = TilingScheme(
    num_level_zero_tiles=(1, 1),
    crs_name=WEB_MERCATOR_CRS_NAME,
    map_height=EARTH_CIRCUMFERENCE_WGS84
)

TilingScheme.GEOGRAPHIC = TilingScheme(
    num_level_zero_tiles=(2, 1),
    crs_name=GEOGRAPHIC_CRS_NAME,
    map_height=180.
)


def get_unit_factor(unit_name_from: str, unit_name_to: str) -> float:
    """
    Get the factor to convert from one unit into another
    with units given by *unit_name_from* and *unit_name_to*.
    """
    from_meter = _is_meter_unit(unit_name_from)
    from_degree = _is_degree_unit(unit_name_from)
    if not from_meter and not from_degree:
        raise ValueError(f'unsupported units {unit_name_from!r}.'
                         f' Unit must be either meters or degrees.')

    to_meter = _is_meter_unit(unit_name_to)
    to_degree = _is_degree_unit(unit_name_to)
    if not to_meter and not to_degree:
        raise ValueError(f'unsupported units {unit_name_from!r}.'
                         f' Unit must be either meters or degrees.')

    if from_meter and to_degree:
        return 360 / EARTH_CIRCUMFERENCE_WGS84
    if from_degree and to_meter:
        return EARTH_CIRCUMFERENCE_WGS84 / 360
    return 1.


def subdivide_size(size: Tuple[int, int],
                   tile_size: Tuple[int, int]) -> List[Tuple[int, int]]:
    x_size, y_size = size
    tile_size_x, tile_size_y = tile_size
    sizes = [(x_size, y_size)]
    while True:
        if x_size <= tile_size_x or y_size <= tile_size_y:
            break
        x_size = (x_size + 1) // 2
        y_size = (y_size + 1) // 2
        sizes.append((x_size, y_size))
    return sizes


def get_num_levels(size: Tuple[int, int],
                   tile_size: Tuple[int, int]) -> int:
    return len(subdivide_size(size, tile_size))


def _is_meter_unit(unit_name: str) -> bool:
    return unit_name.lower() in ('m',
                                 'metre', 'metres',
                                 'meter', 'meters')


def _is_degree_unit(unit_name: str) -> bool:
    return unit_name.lower() in ('°', 'deg',
                                 'degree', 'degrees',
                                 'decimal_degree', 'decimal_degrees')



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from typing import List, Callable, Sequence, Optional, Tuple

import xarray as xr
from deprecated import deprecated

from xcube.core.dsio import open_dataset

PyramidLevelCallback = Callable[[xr.Dataset, int, int], Optional[xr.Dataset]]

_DEPRECATED_READ = (
    'No longer in use. To read multi-level datasets, use:\n'
    '>>> data_store = new_data_store("file")\n'
    '>>> ml_dataset = data_store.open_data(path + ".levels")'
)
_DEPRECATED_WRITE = (
    'No longer in use. To write multi-level datasets, use:\n'
    '>>> data_store = new_data_store("file")\n'
    '>>> data_store.write_data(dataset, path + ".levels")'
)


@deprecated(version='0.10.2', reason=_DEPRECATED_WRITE)
def compute_levels(
        dataset: xr.Dataset,
        spatial_dims: Tuple[str, str] = None,
        spatial_shape: Tuple[int, int] = None,
        spatial_tile_shape: Tuple[int, int] = None,
        var_names: Sequence[str] = None,
        num_levels_max: int = None,
        post_process_level: PyramidLevelCallback = None,
        progress_monitor: PyramidLevelCallback = None
) -> List[xr.Dataset]:
    """
    Transform the given *dataset* into the levels of a multi-level
    pyramid with spatial resolution decreasing by a factor of two
    in both spatial dimensions.

    It is assumed that the spatial dimensions of each variable are
    the inner-most, that is, the last two elements of a variable's
    shape provide the spatial dimension sizes.

    :param dataset: The input dataset to be turned into
        a multi-level pyramid.
    :param spatial_dims: If given, only variables are considered
        whose last to dimension elements match the given *spatial_dims*.
    :param spatial_shape: If given, only variables are
        considered whose last to shape elements match the
        given *spatial_shape*.
    :param spatial_tile_shape: If given, chunking will match the
        provided *spatial_tile_shape*.
    :param var_names: Variables to consider. If None, all variables
        with at least two dimensions are considered.
    :param num_levels_max: If given, the maximum number
        of pyramid levels.
    :param post_process_level: If given, the function will be
        called for each level and must return a dataset.
    :param progress_monitor: If given, the function will be
        called for each level.
    :return: A list of dataset instances representing
        the multi-level pyramid.
    """
    dropped_vars, spatial_shape, spatial_tile_shape = \
        _filter_level_source_dataset(dataset, var_names, spatial_dims,
                                     spatial_shape, spatial_tile_shape)
    if dropped_vars:
        dataset = dataset.drop_vars(dropped_vars)

    if not tuple(dataset.data_vars):
        raise ValueError(
            "cannot compute pyramid levels because "
            "no suitable data variables were found"
        )

    if spatial_tile_shape is None:
        tile_w = min(spatial_shape[0], 512)
        tile_h = min(spatial_shape[1], 512)
        spatial_tile_shape = tile_w, tile_h

    # Count num_levels
    level_shapes = _compute_level_shapes(spatial_shape,
                                         spatial_tile_shape,
                                         num_levels_max=num_levels_max)
    num_levels = len(level_shapes)

    # Compute levels
    level_dataset = dataset
    level_datasets = []
    for level in range(num_levels):
        if level > 0:
            # Down-sample levels
            downsampled_vars = {}
            for var_name in level_dataset.data_vars:
                var = level_dataset.data_vars[var_name]
                # For time being, we use the simplest and
                # likely fastest downsampling I can think of
                downsampled_var = var[..., ::2, ::2]
                if downsampled_var.shape[-2:] != level_shapes[level]:
                    import warnings
                    warnings.warn(
                        f"unexpected spatial shape for down-sampled"
                        f" variable {var_name!r}:"
                        f" expected {level_shapes[level]},"
                        f" but found {downsampled_var.shape[-2:]}"
                    )
                downsampled_vars[var_name] = downsampled_var
            level_dataset = xr.Dataset(downsampled_vars,
                                       attrs=level_dataset.attrs)

        level_dataset = _tile_level_dataset(level_dataset,
                                            spatial_tile_shape)

        # Apply post processor, if any
        if post_process_level is not None:
            level_dataset = post_process_level(level_dataset,
                                               len(level_datasets),
                                               num_levels)

        # Inform progress monitor, if any
        if progress_monitor is not None:
            progress_monitor(level_dataset, len(level_datasets),
                             num_levels)

        # Collect level dataset
        level_datasets.append(level_dataset)

    return level_datasets


@deprecated(version='0.10.2', reason=_DEPRECATED_WRITE)
def write_levels(output_path: str,
                 dataset: xr.Dataset = None,
                 input_path: str = None,
                 link_input: bool = False,
                 progress_monitor: PyramidLevelCallback = None,
                 **kwargs) -> List[xr.Dataset]:
    """
    Transform the given dataset given by a *dataset* instance
    or *input_path* string into the levels of a multi-level pyramid
    with spatial resolution decreasing by a factor of two in both
    spatial dimensions and write them to *output_path*.

    One of *dataset* and *input_path* must be given.

    :param output_path: Output path
    :param dataset: Dataset to be converted and written as levels.
    :param input_path: Input path to a dataset to be transformed
        and written as levels.
    :param link_input: Just link the dataset at level zero
        instead of writing it.
    :param progress_monitor: An optional progress monitor.
    :param kwargs: Keyword-arguments accepted by the
        ``compute_levels()`` function.
    :return: A list of dataset instances representing
        the multi-level pyramid.
    """
    if dataset is None and input_path is None:
        raise ValueError(
            "at least one of dataset or input_path must be given"
        )

    if link_input and input_path is None:
        raise ValueError(
            "input_path must be provided to link input"
        )

    _post_process_level = kwargs.pop("post_process_level") \
        if "post_process_level" in kwargs else None

    def post_process_level(level_dataset, index, num_levels):
        if _post_process_level is not None:
            level_dataset = _post_process_level(level_dataset,
                                                index,
                                                num_levels)

        if index == 0 and link_input:
            with open(os.path.join(output_path, f"{index}.link"), "w") as fp:
                fp.write(input_path)
        else:
            path = os.path.join(output_path, f"{index}.zarr")
            level_dataset.to_zarr(path)
            level_dataset.close()
            level_dataset = xr.open_zarr(path)

        if progress_monitor is not None:
            progress_monitor(level_dataset, index, num_levels)

        return level_dataset

    if not os.path.exists(output_path):
        os.makedirs(output_path)

    if dataset is None:
        dataset = open_dataset(input_path)

    return compute_levels(dataset,
                          post_process_level=post_process_level,
                          **kwargs)


@deprecated(version='0.10.2', reason=_DEPRECATED_READ)
def read_levels(
        dir_path: str,
        progress_monitor: PyramidLevelCallback = None
) -> List[xr.Dataset]:
    """
    Read the of a multi-level pyramid with spatial resolution
    decreasing by a factor of two in both spatial dimensions.

    :param dir_path: The directory path.
    :param progress_monitor: An optional progress monitor.
    :return: A list of dataset instances representing
        the multi-level pyramid.
    """
    file_paths = os.listdir(dir_path)
    level_paths = {}
    num_levels = -1
    for filename in file_paths:
        file_path = os.path.join(dir_path, filename)
        basename, ext = os.path.splitext(filename)
        if basename.isdigit():
            index = int(basename)
            num_levels = max(num_levels, index + 1)
            if os.path.isfile(file_path) and ext == ".link":
                level_paths[index] = (ext, file_path)
            elif os.path.isdir(file_path) and ext == ".zarr":
                level_paths[index] = (ext, file_path)

    if num_levels != len(level_paths):
        raise ValueError(f"Inconsistent pyramid directory:"
                         f" expected {num_levels} but found"
                         f" {len(level_paths)} entries:"
                         f" {dir_path}")

    levels = []
    for index in range(num_levels):
        ext, file_path = level_paths[index]
        if ext == ".link":
            with open(file_path, "r") as fp:
                link_file_path = fp.read()
            if not os.path.isabs(link_file_path):
                parent_dir_path = os.path.abspath(
                    os.path.dirname(dir_path) or '.'
                )
                link_file_path = os.path.join(
                    parent_dir_path, link_file_path
                )
            dataset = xr.open_zarr(link_file_path)
        else:
            dataset = xr.open_zarr(file_path)
        if progress_monitor is not None:
            progress_monitor(dataset, index, num_levels)
        levels.append(dataset)
    return levels


def _tile_chunk(size, tile_size):
    last_tile_size = size % tile_size
    if last_tile_size != 0:
        return (tile_size,) * (size // tile_size) + (last_tile_size,)
    return tile_size


def _tile_level_dataset(level_dataset, spatial_tile_shape):
    tile_height, tile_width = spatial_tile_shape

    # Chunk variables in level dataset according to spatial_tile_shape
    chunked_vars = {}

    # Chunk data variables according to tile size
    for var_name in level_dataset.data_vars:
        var = level_dataset.data_vars[var_name]
        height, width = var.shape[-2:]
        zarr_chunks = (1,) * (var.ndim - 2) + (
            tile_height, tile_width
        )
        dask_chunks = (1,) * (var.ndim - 2) + (
            _tile_chunk(height, tile_height), _tile_chunk(width, tile_width)
        )
        dask_chunks = {var.dims[i]: dask_chunks[i] for i in range(var.ndim)}
        chunked_var = var.chunk(chunks=dask_chunks)
        chunked_var.encoding.update(chunks=zarr_chunks)
        chunked_vars[var_name] = chunked_var

    # Make coordinate variable chunks equal to their shape
    # TODO (forman): find out if chunking the spatial coordinates
    #  according to tile size improves performance
    for var_name in level_dataset.coords:
        var = level_dataset.coords[var_name]
        zarr_chunks = var.shape
        dask_chunks = {var.dims[i]: var.shape[i] for i in range(var.ndim)}
        chunked_var = var.chunk(chunks=dask_chunks)
        chunked_var.encoding.update(chunks=zarr_chunks)
        chunked_vars[var_name] = chunked_var

    return level_dataset.assign(variables=chunked_vars)


def _compute_level_shapes(
        spatial_shape,
        spatial_tile_shape,
        num_levels_max=None
) -> List[Tuple[int, int]]:
    height, width = spatial_shape
    tile_height, tile_width = spatial_tile_shape
    num_levels_max = num_levels_max or -1
    level_shapes = [(height, width)]
    while True:
        width = (width + 1) // 2
        height = (height + 1) // 2
        if width < tile_width \
                or height < tile_height \
                or num_levels_max == len(level_shapes):
            break
        level_shapes.append((height, width))
    return level_shapes


def _filter_level_source_dataset(dataset,
                                 var_names=None,
                                 spatial_dims=None,
                                 spatial_shape=None,
                                 spatial_tile_shape=None):
    if var_names:
        var_names = set(var_names)
        dropped_vars = list(set(dataset.data_vars).difference(var_names))
    else:
        var_names = set(dataset.data_vars)
        dropped_vars = []

    # Collect data variables to be dropped,
    # derive missing information from spatial data variables
    for var_name in var_names:

        if var_name not in dataset.data_vars:
            raise ValueError(f"variable {var_name} not found")

        var = dataset[var_name]

        if var.ndim < 2:
            # Must have at least the two spatial dimensions
            dropped_vars.append(var_name)
            continue

        if spatial_dims is None:
            spatial_dims = var.dims[-2:]
        elif spatial_dims != var.dims[-2:]:
            # Spatial dimensions don't fit
            dropped_vars.append(var_name)
            continue

        if spatial_shape is None:
            spatial_shape = var.shape[-2:]
        elif spatial_shape != var.shape[-2:]:
            # Spatial dimension sizes don't fit
            dropped_vars.append(var_name)
            continue

        if spatial_tile_shape is None and var.chunks is not None:
            def chunk_to_int(chunk):
                return chunk if isinstance(chunk, int) else max(chunk)

            spatial_tile_shape = tuple(map(chunk_to_int, var.chunks[-2:]))

    return dropped_vars, spatial_shape, spatial_tile_shape



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import abc
import importlib
import pkgutil
import sys
import time
import traceback
import warnings
from typing import Callable, Dict, Optional, Any, List

from pkg_resources import iter_entry_points

from xcube.constants import PLUGIN_ENTRY_POINT_GROUP_NAME
from xcube.constants import PLUGIN_INIT_TIME__WARN_LIMIT
from xcube.constants import PLUGIN_LOAD_TIME_WARN_LIMIT
from xcube.constants import PLUGIN_MODULE_FUNCTION_NAME
from xcube.constants import PLUGIN_MODULE_NAME
from xcube.constants import PLUGIN_MODULE_PREFIX
from xcube.util.extension import Extension
from xcube.util.extension import ExtensionRegistry

_PLUGIN_REGISTRY_INIT = dict()

# Mapping of xcube entry point names
# to JSON-serializable plugin meta-information.
_PLUGIN_REGISTRY: Dict[str, Dict[str, Any]] = _PLUGIN_REGISTRY_INIT


def init_plugins() -> None:
    """Load plugins if not already done."""
    global _PLUGIN_REGISTRY
    if _PLUGIN_REGISTRY is _PLUGIN_REGISTRY_INIT:
        _PLUGIN_REGISTRY = load_plugins()


def get_plugins() -> Dict[str, Dict]:
    """Get mapping of "xcube_plugins" entry point names
     to JSON-serializable plugin meta-information.
     """
    init_plugins()
    global _PLUGIN_REGISTRY
    return dict(_PLUGIN_REGISTRY)


def get_extension_registry() -> ExtensionRegistry:
    """Get populated extension registry."""
    from xcube.util.extension import get_extension_registry
    init_plugins()
    return get_extension_registry()


def discover_plugin_modules(module_prefixes=None):
    module_prefixes = module_prefixes or [PLUGIN_MODULE_PREFIX]
    entry_points = []
    for module_finder, module_name, ispkg in pkgutil.iter_modules():
        if any([module_name.startswith(module_prefix)
                for module_prefix in module_prefixes]):
            # Note: Consider turning this into debug log,
            #  but logging is not yet configured at this point.
            # print(f'xcube plugin module found: {module_name}')
            entry_points.append(_ModuleEntryPoint(module_name))
    return entry_points


def load_plugins(
        entry_points=None,
        ext_registry: Optional[ExtensionRegistry] = None
) -> Dict[str, Dict[str, Any]]:
    if ext_registry is None:
        from xcube.util.extension import get_extension_registry
        ext_registry = get_extension_registry()

    plugins = {}

    if entry_points:
        _collect_plugins(
            entry_points,
            ext_registry,
            True,
            plugins
        )
    else:
        # verbose=True for specified xcube plugins.
        _collect_plugins(
            list(iter_entry_points(group=PLUGIN_ENTRY_POINT_GROUP_NAME,
                                   name=None)),
            ext_registry,
            True,
            plugins
        )
        # verbose=False for auto-detected xcube plugins,
        # where package name starts with "xcube_" but an
        # entry point is not explicitly specified.
        _collect_plugins(
            discover_plugin_modules(),
            ext_registry,
            False,
            plugins
        )

    return plugins


def _collect_plugins(entry_points: List[Any],
                     ext_registry: ExtensionRegistry,
                     verbose: bool,
                     plugins: Dict[str, Dict[str, Any]]):
    for entry_point in entry_points:
        # Note: Consider turning this into debug log,
        #  but logging is not yet configured at this point.
        #  print(f'loading xcube plugin {entry_point.name!r}')

        t0 = time.perf_counter()

        # noinspection PyBroadException
        try:
            plugin_init_function = entry_point.load()
        except Exception as e:
            if verbose:
                _emit_warning_for_error(entry_point, e)
            continue

        if plugin_init_function is None:
            # Not a plugin
            continue

        if verbose:
            millis = int(1000 * (time.perf_counter() - t0))
            if millis >= PLUGIN_LOAD_TIME_WARN_LIMIT:
                _emit_warning_for_slow_load(entry_point, millis)

        if not callable(plugin_init_function):
            if verbose:
                _emit_warning_on_init_function(entry_point,
                                               plugin_init_function)
            continue

        t0 = time.perf_counter()

        # noinspection PyBroadException
        try:
            plugin_init_function(ext_registry)
        except Exception as e:
            if verbose:
                _emit_warning_for_error(entry_point, e)
            continue

        if verbose:
            millis = int(1000 * (time.perf_counter() - t0))
            if millis >= PLUGIN_INIT_TIME__WARN_LIMIT:
                _emit_warning_for_slow_load(entry_point, millis)

        plugins[entry_point.name] = {'name': entry_point.name,
                                     'doc': plugin_init_function.__doc__}


def _emit_warning_on_init_function(entry_point, plugin_init_function):
    # We use warning and not raise to allow loading
    # xcube despite a broken plugin. Raise would stop xcube.
    warnings.warn(f'xcube plugin {entry_point.name!r}'
                  f' must be callable'
                  f' but got a {type(plugin_init_function)!r}')


def _emit_warning_for_slow_load(entry_point, millis):
    warnings.warn(
        f'Initializing xcube plugin {entry_point.name!r}'
        f' took {millis} ms,'
        f' consider code optimization.'
        f' (For example, avoid eager import of packages,'
        f' consider lazy loading of resources, etc.)'
    )


def _emit_warning_for_error(entry_point, e):
    # We use warning and not raise to allow loading
    # xcube despite a broken plugin. Raise would stop xcube.
    warnings.warn(f'Unexpected exception while loading'
                  f' xcube plugin {entry_point.name!r}: {e}')
    traceback.print_exc(file=sys.stderr)


class _ModuleEntryPoint:
    def __init__(self, module_name: str):
        self._module_name = module_name

    @property
    def name(self) -> str:
        return self._module_name

    def load(self) -> Optional[Callable]:
        """
        Load function "init_plugin()" either from "<module_name>.plugin" or "<module_name>"
        :return: plugin init function
        """
        try:
            plugin_module = importlib.import_module(
                f'{self._module_name}.{PLUGIN_MODULE_NAME}'
            )
        except ModuleNotFoundError:
            plugin_module = importlib.import_module(self._module_name)

        module_func_name = PLUGIN_MODULE_FUNCTION_NAME
        if hasattr(plugin_module, module_func_name):
            module_func = getattr(plugin_module, module_func_name)
            if callable(module_func):
                return module_func

        # Here: The module name looks like an xcube-plugin
        # but lacks a callable named module_func_name

        return None


class ExtensionComponent(metaclass=abc.ABCMeta):
    """
    Utility base class for extension components.
    """

    def __init__(self, extension_point: str, name: str):
        if not extension_point:
            raise ValueError('extension_point must be given')
        if not name:
            raise ValueError('name must be given')
        self._extension_point = extension_point
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    def extension_point(self) -> str:
        """
        :return: The extension point for this component.
        """
        return self._extension_point

    @property
    def extension(self) -> Optional[Extension]:
        """
        :return: The extension for this component.
            None, if it has not (yet) been registered as an extension.
        """
        return get_extension_registry().get_extension(
            self._extension_point, self._name
        )

    def get_metadata_attr(self, key: str, default: Any = None) -> Any:
        """
        :return: A metadata attribute for *key* and given *default* value.
        """
        extension = self.extension
        return extension.metadata.get(key, default) \
            if extension is not None else ''



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import os.path
import time
from typing import Type, TypeVar, Dict, Any, Optional, List, Tuple, Union

import requests

from xcube.util.assertions import assert_instance
from xcube.util.jsonschema import JsonObject
from xcube.util.progress import observe_progress
from .config import ServiceConfig
from .response import CubeGeneratorProgress
from .response import CubeGeneratorState
from .response import CubeGeneratorToken
from .response import CubeInfoWithCostsResult
from ..error import CubeGeneratorError
from ..generator import CubeGenerator
from ..request import CubeGeneratorRequest
from ..request import CubeGeneratorRequestLike
from ..response import CubeGeneratorResult
from ..response import CubeInfoResult
from ..response import GenericCubeGeneratorResult

_BASE_HEADERS = {
    "Accept": "application/json",
    # "Content-Type": "application/json",
}

R = TypeVar('R', bound=JsonObject)


class RemoteCubeGenerator(CubeGenerator):
    """
    A cube generator that uses a remote cube generator remote.

    Creates cube views from one or more cube stores, resamples them to a
    common grid, optionally performs some cube transformation, and writes
    the resulting cube to some target cube store.

    :param service_config: An remote configuration object.
    :param verbosity: Level of verbosity, 0 means off.
    :param raise_on_error: Whether to raise a CubeGeneratorError
        exception on generator failures. If False, the default,
        the returned result will have the "status" field set to "error"
        while other fields such as "message", "traceback", "output"
        provide more failure details.
    """

    def __init__(self,
                 service_config: ServiceConfig,
                 progress_period: float = 1.0,
                 raise_on_error: bool = False,
                 verbosity: int = 0):
        super().__init__(raise_on_error=raise_on_error,
                         verbosity=verbosity)
        assert_instance(service_config, ServiceConfig, 'service_config')
        assert_instance(progress_period, (int, float), 'progress_period')
        self._service_config: ServiceConfig = service_config
        self._access_token: Optional[str] = service_config.access_token
        self._progress_period: float = progress_period

    def endpoint_op(self, op_path: str) -> str:
        return f'{self._service_config.endpoint_url}{op_path}'

    @property
    def auth_headers(self) -> Dict:
        access_token = self.access_token
        if access_token is not None:
            return {
                **_BASE_HEADERS,
                'Authorization': f'Bearer {self.access_token}',
            }
        return dict(_BASE_HEADERS)

    @property
    def access_token(self) -> Optional[str]:
        if self._access_token is None:
            if self._service_config.client_id is None \
                    and self._service_config.client_secret is None:
                return None

            request_data = {
                "audience": self._service_config.endpoint_url,
                "client_id": self._service_config.client_id,
                "client_secret": self._service_config.client_secret,
                "grant_type": "client-credentials",
            }
            response = requests.post(self.endpoint_op('oauth/token'),
                                     json=request_data,
                                     headers=_BASE_HEADERS)
            token_response: CubeGeneratorToken = \
                self._parse_response(response,
                                     CubeGeneratorToken,
                                     request_data=request_data)
            self._access_token = token_response.access_token
        return self._access_token

    def _get_cube_info(self, request: CubeGeneratorRequestLike) \
            -> CubeInfoResult:
        request = CubeGeneratorRequest.normalize(request).for_service()
        request_data = request.to_dict()
        response = requests.post(self.endpoint_op('cubegens/info'),
                                 json=request_data,
                                 headers=self.auth_headers)
        response_type = CubeInfoWithCostsResult \
            if self._access_token else CubeInfoResult
        return self._parse_response(response,
                                    response_type=response_type,
                                    request_data=request_data)

    def _generate_cube(self, request: CubeGeneratorRequestLike) \
            -> CubeGeneratorResult:
        request = CubeGeneratorRequest.normalize(request).for_service()

        response = self._submit_gen_request(request)
        cubegen_id, result, _ = \
            self._get_cube_generator_result(response)
        if result is not None:
            return result

        last_worked = 0
        with observe_progress('Generating cube', 100) as cm:
            while True:
                time.sleep(self._progress_period)

                response = requests.get(
                    self.endpoint_op(f'cubegens/{cubegen_id}'),
                    headers=self.auth_headers
                )
                _, result, progress = \
                    self._get_cube_generator_result(response)
                if result is not None:
                    return result

                if progress is not None and len(progress) > 0:
                    progress_state = progress[0].state
                    total_work = progress_state.total_work
                    progress = progress_state.progress or 0
                    worked = progress * total_work
                    work = 100 * ((worked - last_worked) / total_work)
                    if work > 0:
                        cm.worked(work)
                        last_worked = worked

    def _submit_gen_request(self, request: CubeGeneratorRequest):
        request_dict = request.to_dict()

        user_code_path = request_dict \
            .get('code_config', {}) \
            .get('file_set', {}) \
            .get('path')

        if user_code_path:
            user_code_filename = os.path.basename(user_code_path)
            request_dict['code_config']['file_set']['path'] \
                = user_code_filename
            return requests.put(
                self.endpoint_op('cubegens/code'),
                headers=self.auth_headers,
                files={
                    'body': (
                        'request.json',
                        json.dumps(request_dict, indent=2),
                        'application/json'
                    ),
                    'user_code': (
                        user_code_filename,
                        open(user_code_path, 'rb'),
                        'application/octet-stream'
                    )
                }
            )
        else:
            return requests.put(
                self.endpoint_op('cubegens'),
                json=request_dict,
                headers=self.auth_headers
            )

    @staticmethod
    def _get_data_id(request: CubeGeneratorRequest, default: str) -> CubeGeneratorResult:
        data_id = request.output_config.data_id
        return data_id if data_id else default

    def _get_cube_generator_result(
            self,
            response: requests.Response,
            request_data: Dict[str, Any] = None
    ) -> Tuple[str,
               Optional[CubeGeneratorResult],
               Optional[List[CubeGeneratorProgress]]]:
        state = self._get_cube_generator_state(response, request_data)
        result = None
        if state.job_status.succeeded:
            result = state.job_result
            if isinstance(result, CubeGeneratorResult):
                status_code = result.status_code or response.status_code
                result = result.derive(status_code=status_code,
                                       output=state.output)
            else:
                result = self._new_invalid_result(state)
        elif state.job_status.failed:
            result = state.job_result
            if isinstance(result, CubeGeneratorResult):
                status_code = result.status_code or response.status_code
                result = result.derive(status='error',
                                       status_code=status_code,
                                       output=state.output)
            else:
                result = self._new_invalid_result(state)
        return state.job_id, result, state.progress

    def _get_cube_generator_state(
            self,
            response: requests.Response,
            request_data: Dict[str, Any] = None
    ) -> CubeGeneratorState:
        return self._parse_response(response,
                                    CubeGeneratorState,
                                    request_data=request_data)

    def _parse_response(self,
                        response: requests.Response,
                        response_type: Type[R],
                        request_data: Dict[str, Any] = None) -> R:

        # noinspection PyBroadException
        try:
            response_data = response.json()
        except BaseException as e:
            raise self._new_response_error(response, msg=e) from e

        if self._verbosity >= 3:
            self.__dump_json(response.request.method,
                             response.url,
                             request_data,
                             response_data)

        if response_data is None:
            raise self._new_response_error(
                response, msg='no response'
            )

        if not isinstance(response_data, dict):
            raise self._new_response_error(
                response, msg='response must be a dictionary'
            )

        # noinspection PyBroadException
        try:
            result = response_type.from_dict(response_data)
        except BaseException as e:
            raise self._new_response_error(
                response,
                msg=f'failed parsing response: {e}'
            ) from e

        if isinstance(result, GenericCubeGeneratorResult) \
                and result.status_code is None:
            result = result.derive(status_code=response.status_code)

        return result

    @classmethod
    def _new_response_error(
            cls,
            response: requests.Response,
            msg: Union[str, BaseException]
    ) -> CubeGeneratorError:
        try:
            response.raise_for_status()
        except requests.HTTPError as http_error:
            return CubeGeneratorError(
                f'{msg}: {http_error}',
                status_code=response.status_code
            )
        return cls._new_unexpected_response_error(response, msg=msg)

    @classmethod
    def _new_unexpected_response_error(
            cls,
            response: requests.Response,
            msg: Union[str, BaseException]
    ) -> CubeGeneratorError:
        return CubeGeneratorError(f'Client error: unexpected response'
                                  f' from API call {response.url},'
                                  f' status code {response.status_code}:'
                                  f' {msg}',
                                  status_code=422)

    @classmethod
    def _new_invalid_result(cls, state: CubeGeneratorState):
        return CubeGeneratorResult(
            status='error',
            status_code=422,
            message='missing cube generator result',
            output=state.output
        )

    @classmethod
    def __dump_json(cls, method, url, request_data, response_data):
        """
        Dump debug info as JSON to stdout.

        Used for debugging only.
        """
        url_line = f'{method} {url}:'
        request_line = 'Request:'
        response_line = 'Response:'

        print('=' * len(url_line))
        print(url_line)
        print('=' * len(url_line))
        print('-' * len(request_line))
        print(request_line)
        print('-' * len(request_line))
        print(json.dumps(request_data, indent=2))
        print('-' * len(response_line))
        print(response_line)
        print('-' * len(response_line))
        print(json.dumps(response_data, indent=2))



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Union, Tuple

import numpy as np
import pyproj
import pyproj.transformer as pt
import xarray as xr

from .base import DEFAULT_TOLERANCE
from .base import GridMapping
from .coords import new_grid_mapping_from_coords
from .helpers import _assert_valid_xy_names
from .helpers import _normalize_crs


# Cannot be used, but should, see TODO in transform_grid_mapping()
#
# class TransformedGridMapping(GridMapping, abc.ABC):
#     """
#     Grid mapping constructed from 1D/2D coordinate variables and a CRS.
#     """
#
#     def __init__(self,
#                  /,
#                  xy_coords: xr.DataArray,
#                  **kwargs):
#         self._xy_coords = xy_coords
#         super().__init__(**kwargs)
#
#     @property
#     def xy_coords(self) -> xr.DataArray:
#         return self._xy_coords
#

def transform_grid_mapping(
        grid_mapping: GridMapping,
        crs: Union[str, pyproj.crs.CRS],
        *,
        tile_size: Union[int, Tuple[int, int]] = None,
        xy_var_names: Tuple[str, str] = None,
        tolerance: float = DEFAULT_TOLERANCE,
) -> GridMapping:
    target_crs = _normalize_crs(crs)

    if xy_var_names:
        _assert_valid_xy_names(xy_var_names, name='xy_var_names')

    source_crs = grid_mapping.crs
    if source_crs == target_crs:
        if tile_size is not None or xy_var_names is not None:
            return grid_mapping.derive(tile_size=tile_size,
                                       xy_var_names=xy_var_names)
        return grid_mapping

    transformer = pt.Transformer.from_crs(source_crs,
                                          target_crs,
                                          always_xy=True)

    def _transform(block: np.ndarray) -> np.ndarray:
        x1, y1 = block
        x2, y2 = transformer.transform(x1, y1)
        return np.stack([x2, y2])

    xy_coords = xr.apply_ufunc(_transform,
                               grid_mapping.xy_coords,
                               output_dtypes=[np.float64],
                               dask='parallelized')

    xy_var_names = xy_var_names or ('transformed_x', 'transformed_y')

    # TODO: Use a specialized grid mapping here that can store the
    #   *xy_coords* directly. Splitting the xy_coords dask array into
    #   x,y components as done here may be very inefficient for larger
    #   arrays, because x cannot be computed independently from y.
    #   This means, any access of x chunks will cause y chunks to be
    #   computed too and vice versa. As same operations are performed
    #   on x and y arrays, this will take twice as long as if operation
    #   would be performed on the xy_coords dask array directly.

    return new_grid_mapping_from_coords(
        x_coords=xr.DataArray(xy_coords[0], name=xy_var_names[0]),
        y_coords=xr.DataArray(xy_coords[1], name=xy_var_names[1]),
        crs=target_crs,
        tile_size=tile_size,
        tolerance=tolerance,
    )



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, List

from xcube.util.assertions import assert_instance


class CubeGeneratorError(ValueError):
    """
    Represent a client or server error that
    may occur in the data cube generators.

    :param args: arguments passed to base exceptions.
    :param status_code: Optional status code of the error (an integer).
        If given, HTTP error codes should be used.
        The range 400-499 indicates client errors.
        The range 500-599 indicates server errors.
    :param remote_traceback: Traceback of an error
        occurred in a remote process.
    :param remote_output:  Terminal output of a remote process.
    """

    def __init__(self,
                 *args,
                 status_code: Optional[int] = None,
                 remote_traceback: Optional[List[str]] = None,
                 remote_output: Optional[List[str]] = None,
                 **kwargs):
        # noinspection PyArgumentList
        super().__init__(*args, **kwargs)
        if status_code is not None:
            assert_instance(status_code, int, 'status_code')
        self._status_code = status_code
        self._remote_traceback = remote_traceback
        self._remote_output = remote_output

    @property
    def status_code(self) -> Optional[int]:
        """
        Status code of the error.
        May be None.
        """
        return self._status_code

    @property
    def remote_traceback(self) -> Optional[List[str]]:
        """
        Traceback of an error occurred in a remote process.
        May be None.
        """
        return self._remote_traceback

    @property
    def remote_output(self) -> Optional[List[str]]:
        """
        Terminal output of a remote process.
        May be None.
        """
        return self._remote_output



# The MIT License (MIT)
# Copyright (c) 2020 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import collections.abc
from abc import ABC, abstractmethod
from typing import Dict, Any, Callable, Mapping, Sequence, Union, Tuple, \
    Optional

from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true

# Make sure rfc3339-validator package is installed : jsonschema uses it for
# validating instances of JsonDateSchema and JsonDatetimeSchema.
# Use of __import__ avoids "unused package" warnings.
__import__('rfc3339_validator')

import jsonschema

from xcube.util.ipython import register_json_formatter
from xcube.util.undefined import UNDEFINED

Factory = Callable[..., Any]
Serializer = Callable[..., Any]

_TYPES_ENUM = {'null', 'boolean', 'integer', 'number', 'string', 'array',
               'object'}
_NUMERIC_TYPES_ENUM = {'integer', 'number'}


class JsonSchema(ABC):

    # noinspection PyShadowingBuiltins
    def __init__(self,
                 type: Union[str, Sequence[str]] = None,
                 default: Any = UNDEFINED,
                 const: Any = UNDEFINED,
                 enum: Sequence[Any] = None,
                 nullable: bool = None,
                 title: str = None,
                 description: str = None,
                 examples: str = None,
                 factory: Factory = None,
                 serializer: Serializer = None):
        if type is not None and type not in _TYPES_ENUM:
            names = ', '.join(map(lambda t: f'"{t}"', sorted(list(_TYPES_ENUM))))
            raise ValueError(f'type must be one of {names}')
        if factory is not None and not callable(factory):
            raise ValueError('factory must be callable')
        if serializer is not None and not callable(serializer):
            raise ValueError('serializer must be callable')
        self.type: Optional[Union[str, Sequence[str]]] = type
        self.default: Any = default
        self.const: Any = const
        self.enum: Optional[Any] = list(enum) if enum is not None else None
        self.nullable: Optional[bool] = nullable
        self.title: Optional[str] = title
        self.description: Optional[str] = description
        self.examples: Optional[str] = examples
        self.factory = factory
        self.serializer = serializer

    def to_dict(self) -> Dict[str, Any]:
        d = dict()
        if self.type is not None:
            if self.nullable is True and self.type != 'null':
                d.update(type=[self.type, 'null'])
            else:
                d.update(type=self.type)
        elif self.nullable is True:
            d.update(type='null')
        if self.default != UNDEFINED:
            d.update(default=self.default)
        if self.const != UNDEFINED:
            d.update(const=self.const)
        if self.enum is not None:
            d.update(enum=self.enum)
        if self.title is not None:
            d.update(title=self.title)
        if self.description is not None:
            d.update(description=self.description)
        if self.examples is not None:
            d.update(examples=self.examples)
        return d

    def validate_instance(self, instance: Any):
        """Validate JSON value *instance*."""
        # As a base for our custom validator, we hard-code a validator
        # version; JSON metaschema versions are not guaranteed to be backward
        # compatible, so if we use jsonschema.validators.validator_for(
        # schema), our schema may become invalid when the default validator
        # changes. The metaschema version can also be pinned in the schema
        # itself with an entry like this:
        # '$schema': 'http://json-schema.org/draft-07/schema'
        # However, this makes it more work to keep unit tests up to date, and
        # it's fiddly to adapt the recursive to_dict code to make sure that
        # the metaschema key *only* appears in the outermost dictionary.
        base_validator = jsonschema.validators.Draft7Validator

        # By default, jsonschema only recognizes lists as arrays. Here we derive
        # and use a custom validator which recognizes both lists and tuples as
        # arrays.
        new_type_checker = \
            base_validator.TYPE_CHECKER.redefine(
                'array',
                lambda checker, inst: isinstance(inst, (list, tuple)))
        custom_validator = \
            jsonschema.validators.extend(base_validator,
                                         type_checker=new_type_checker)

        # jsconschema needs extra packages installed to validate some formats;
        # if they are missing, the format check will be skipped silently. For
        # date-time format, strict_rfc3339 or rfc3339-validator is required.
        jsonschema.validate(instance=instance, schema=self.to_dict(),
                            cls=custom_validator,
                            format_checker=jsonschema.draft7_format_checker)

    def to_instance(self, value: Any) -> Any:
        """Convert Python object *value* into JSON value and return the validated result."""
        # TODO: support anyOf
        json_instance = self._to_unvalidated_instance(value)
        self.validate_instance(json_instance)
        return json_instance

    def from_instance(self, instance: Any) -> Any:
        """Validate JSON value *instance* and convert it into a Python object."""
        # TODO: support anyOf
        self.validate_instance(instance)
        return self._from_validated_instance(instance)

    @abstractmethod
    def _to_unvalidated_instance(self, value: Any) -> Any:
        """Turn Python object *value* into an unvalidated JSON value."""

    @abstractmethod
    def _from_validated_instance(self, instance: Any) -> Any:
        """Turn validated JSON value *instance* into a Python object."""


class JsonComplexSchema(JsonSchema):
    # TODO: implement JsonComplexTypeSchema more completely
    # For full support one_of, any_of, all_of should also be handled in from_instance and to_instance.

    # noinspection PyShadowingBuiltins
    def __init__(self,
                 one_of: Sequence['JsonSchema'] = None,
                 any_of: Sequence['JsonSchema'] = None,
                 all_of: Sequence['JsonSchema'] = None,
                 **kwargs):
        if len([x for x in (one_of, any_of, all_of) if bool(x)]) != 1:
            raise ValueError('exactly one of one_of, any_of, all_of must be given')
        super().__init__(**kwargs)
        self.one_of = one_of
        self.any_of = any_of
        self.all_of = all_of

    def to_dict(self) -> Dict[str, Any]:
        d = super().to_dict()
        if self.one_of:
            d.update(oneOf=[schema.to_dict() for schema in self.one_of])
        if self.any_of:
            d.update(anyOf=[schema.to_dict() for schema in self.any_of])
        if self.all_of:
            d.update(allOf=[schema.to_dict() for schema in self.all_of])
        return d

    def _to_unvalidated_instance(self, value: Any) -> Any:
        return self.serializer(value) if self.serializer is not None else value

    def _from_validated_instance(self, instance: Any) -> Any:
        return self.factory(instance) if self.factory is not None else instance


class JsonSimpleSchema(JsonSchema):
    # noinspection PyShadowingBuiltins
    def __init__(self, type: str, **kwargs):
        super().__init__(type, **kwargs)
        if not isinstance(type, str):
            raise ValueError(f'illegal type: {type}')

    def _to_unvalidated_instance(self, value: Any) -> Any:
        return self.serializer(value) if self.serializer is not None else value

    def _from_validated_instance(self, instance: Any) -> Any:
        return self.factory(instance) if self.factory is not None else instance


class JsonNullSchema(JsonSimpleSchema):
    def __init__(self, **kwargs):
        super().__init__(type='null', **kwargs)


class JsonBooleanSchema(JsonSimpleSchema):
    def __init__(self, **kwargs):
        super().__init__(type='boolean', **kwargs)


class JsonStringSchema(JsonSimpleSchema):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 format: str = None,
                 pattern: str = None,
                 min_length: int = None,
                 max_length: int = None,
                 **kwargs):
        super().__init__(type='string', **kwargs)
        self.format = format
        self.pattern = pattern
        self.min_length = min_length
        self.max_length = max_length

    def to_dict(self) -> Dict[str, Any]:
        d = super().to_dict()
        if self.format is not None:
            d.update(format=self.format)
        if self.pattern is not None:
            d.update(pattern=self.pattern)
        if self.min_length is not None:
            d.update(minLength=self.min_length)
        if self.max_length is not None:
            d.update(maxLength=self.max_length)
        return d


class JsonDateAndTimeSchemaBase:
    # noinspection PyShadowingBuiltins
    @classmethod
    def _validate_value(cls, value: Optional[str], name: str, format: str):
        if value is not None and \
                not cls._is_valid_value(value, format):
            raise ValueError(f'{name} must be formatted as a "{format}"')

    # noinspection PyShadowingBuiltins
    @classmethod
    def _is_valid_value(cls, value: str, format: str) -> bool:
        try:
            jsonschema.validate(value,
                                dict(type='string', format=format),
                                format_checker=jsonschema.draft7_format_checker)
            return True
        except jsonschema.ValidationError:
            return False


class JsonDateSchema(JsonStringSchema, JsonDateAndTimeSchemaBase):
    """
    JSON schema for date instances.

    :param min_date: optional minimum date.
    :param max_date: optional maximum date.
    """

    # noinspection PyShadowingBuiltins
    def __init__(self,
                 min_date: str = None,
                 max_date: str = None,
                 **kwargs):
        super().__init__(**kwargs, format='date')
        self._validate_value(min_date, 'min_date', 'date')
        self._validate_value(max_date, 'max_date', 'date')
        self.min_date = min_date
        self.max_date = max_date

    def to_dict(self) -> Dict[str, Any]:
        d = super().to_dict()
        if self.min_date is not None:
            d.update(minDate=self.min_date)
        if self.max_date is not None:
            d.update(maxDate=self.max_date)
        return d

    @classmethod
    def new_range(cls,
                  min_date: str = None,
                  max_date: str = None,
                  nullable: bool = False) -> 'JsonArraySchema':
        """
        Return a schema for a date range.
        :param min_date: optional minimum date.
        :param max_date: optional maximum date.
        :param nullable: whether the whole range as well as individual start and end may be None.
        :return: a JsonArraySchema with two items.
        """
        return JsonArraySchema(
            items=[
                JsonDateSchema(min_date=min_date, max_date=max_date, nullable=nullable),
                JsonDateSchema(min_date=min_date, max_date=max_date, nullable=nullable),
            ],
            nullable=nullable)


class JsonDatetimeSchema(JsonStringSchema, JsonDateAndTimeSchemaBase):
    """
    JSON schema for date-time instances.

    :param min_datetime: optional minimum date-time.
    :param max_datetime: optional maximum date-time.
    """

    # noinspection PyShadowingBuiltins
    def __init__(self,
                 min_datetime: str = None,
                 max_datetime: str = None,
                 **kwargs):
        super().__init__(**kwargs, format='date-time')
        self._validate_value(min_datetime, 'min_datetime', 'date-time')
        self._validate_value(max_datetime, 'max_datetime', 'date-time')
        self.min_datetime = min_datetime
        self.max_datetime = max_datetime

    def to_dict(self) -> Dict[str, Any]:
        d = super().to_dict()
        if self.min_datetime is not None:
            d.update(minDatetime=self.min_datetime)
        if self.max_datetime is not None:
            d.update(maxDatetime=self.max_datetime)
        return d

    @classmethod
    def new_range(cls,
                  min_datetime: str = None,
                  max_datetime: str = None,
                  nullable: bool = False) -> 'JsonArraySchema':
        """
        Return a schema for a date-time range.
        :param min_datetime: optional minimum date.
        :param max_datetime: optional maximum date.
        :param nullable: whether the whole range as well as individual start and end dates may be None.
        :return: a JsonArraySchema with two items.
        """
        return JsonArraySchema(items=[
            JsonDatetimeSchema(min_datetime=min_datetime, max_datetime=max_datetime, nullable=nullable),
            JsonDatetimeSchema(min_datetime=min_datetime, max_datetime=max_datetime, nullable=nullable),
        ], nullable=nullable)


class JsonNumberSchema(JsonSimpleSchema):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 type: str = 'number',
                 minimum: Union[int, float] = None,
                 exclusive_minimum: Union[int, float] = None,
                 maximum: Union[int, float] = None,
                 exclusive_maximum: Union[int, float] = None,
                 multiple_of: Union[int, float] = None,
                 **kwargs):
        if type not in _NUMERIC_TYPES_ENUM:
            names = ', '.join(sorted(map(lambda t: f'"{t}"', _NUMERIC_TYPES_ENUM)))
            raise ValueError(f'Type must be one of {names}')
        super().__init__(type, **kwargs)
        self.minimum = minimum
        self.exclusive_minimum = exclusive_minimum
        self.maximum = maximum
        self.exclusive_maximum = exclusive_maximum
        self.multiple_of = multiple_of

    def to_dict(self) -> Dict[str, Any]:
        d = super().to_dict()
        if self.minimum is not None:
            d.update(minimum=self.minimum)
        if self.exclusive_minimum is not None:
            d.update(exclusiveMinimum=self.exclusive_minimum)
        if self.maximum is not None:
            d.update(maximum=self.maximum)
        if self.exclusive_maximum is not None:
            d.update(exclusiveMaximum=self.exclusive_maximum)
        if self.multiple_of is not None:
            d.update(multipleOf=self.multiple_of)
        return d


class JsonIntegerSchema(JsonNumberSchema):
    def __init__(self, **kwargs):
        super().__init__(type='integer', **kwargs)


class JsonArraySchema(JsonSchema):

    def __init__(self,
                 items: Union[JsonSchema, Sequence[JsonSchema]] = None,
                 min_items: int = None,
                 max_items: int = None,
                 unique_items: bool = None,
                 **kwargs):
        super().__init__(type='array', **kwargs)
        self.items = items
        self.min_items = min_items
        self.max_items = max_items
        self.unique_items = unique_items

    def to_dict(self) -> Dict[str, Any]:
        d = super().to_dict()
        if self.items is not None:
            if isinstance(self.items, JsonSchema):
                d.update(items=self.items.to_dict())
            else:
                d.update(items=[item.to_dict() for item in self.items])
        if self.min_items is not None:
            d.update(minItems=self.min_items)
        if self.max_items is not None:
            d.update(maxItems=self.max_items)
        if self.unique_items is not None:
            d.update(uniqueItems=self.unique_items)
        return d

    def _to_unvalidated_instance(self, value: Optional[Sequence[Any]]) -> Optional[Sequence[Any]]:
        if self.serializer:
            return self.serializer(value)
        return self._convert_sequence(value, '_to_unvalidated_instance')

    def _from_validated_instance(self, instance: Sequence[Any]) -> Any:
        obj = self._convert_sequence(instance, '_from_validated_instance')
        return self.factory(obj) if self.factory is not None and obj is not None else obj

    def _convert_sequence(self, sequence: Optional[Sequence[Any]], method_name: str) -> Optional[Sequence[Any]]:
        items_schema = self.items
        if sequence is None:
            # Sequence may be null too
            return None
        if isinstance(items_schema, JsonSchema):
            # Sequence turned into list with items_schema applying to all elements
            return [getattr(items_schema, method_name)(item)
                    for item in sequence]
        if items_schema is not None:
            # Sequence turned into tuple with schema for every position
            return [getattr(items_schema[item_index], method_name)(sequence[item_index])
                    for item_index in range(len(items_schema))]
        # Sequence returned as-is, without schema
        return sequence


class JsonObjectSchema(JsonSchema):

    # TODO: also address property dependencies

    def __init__(self,
                 properties: Mapping[str, JsonSchema] = None,
                 additional_properties: Union[bool, JsonSchema] = None,
                 min_properties: int = None,
                 max_properties: int = None,
                 required: Sequence[str] = None,
                 dependencies: Mapping[str, Union[Sequence[str], JsonSchema]] = None,
                 **kwargs):
        super().__init__(type='object', **kwargs)
        self.properties = dict(properties) if properties else dict()
        self.additional_properties = additional_properties
        self.min_properties = min_properties
        self.max_properties = max_properties
        self.required = list(required) if required else []
        self.dependencies = dict(dependencies) if dependencies else None

    def to_dict(self) -> Dict[str, Any]:
        d = super().to_dict()
        if self.properties is not None:
            d.update(properties={k: v.to_dict() for k, v in self.properties.items()})
        if self.additional_properties is not None:
            d.update(additionalProperties=self.additional_properties.to_dict() \
                if isinstance(self.additional_properties, JsonSchema) else self.additional_properties)
        if self.min_properties is not None:
            d.update(minProperties=self.min_properties)
        if self.max_properties is not None:
            d.update(maxProperties=self.max_properties)
        if self.required:
            d.update(required=list(self.required))
        if self.dependencies:
            d.update(dependencies={k: (v.to_dict() if isinstance(v, JsonSchema) else v)
                                   for k, v in self.dependencies.items()})
        return d

    # TODO: move away. this is a special-purpose utility
    def process_kwargs_subset(self,
                              kwargs: Dict[str, Any],
                              keywords: Sequence[str]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Utility that helps processing keyword-arguments. in *kwargs*:

        Pop every keyword in *keywords* contained in this object schema's properties
        from *kwargs* and put the keyword and value from *kwargs* into a new dictionary.

        Return a 2-element tuple, first *kwargs* with, second the and *kwargs* without the keywords from *keywords*.

        The original *kwargs* is not touched.

        :return a tuple of two new keyword-argument dictionaries
        """
        old_kwargs = dict(kwargs)
        new_kwargs = {}
        for k in keywords:
            if k in old_kwargs:
                new_kwargs[k] = old_kwargs.pop(k)
            elif k in self.properties:
                property_schema = self.properties[k]
                if property_schema.const != UNDEFINED:
                    new_kwargs[k] = property_schema.const
                elif property_schema.default != UNDEFINED and k in self.required:
                    if property_schema.default != UNDEFINED:
                        new_kwargs[k] = property_schema.default
        return new_kwargs, old_kwargs

    def _to_unvalidated_instance(self, value: Any) -> Optional[Mapping[str, Any]]:
        if self.serializer is not None:
            return self.serializer(value)
        return self._convert_mapping(value, '_to_unvalidated_instance')

    def _from_validated_instance(self, instance: Optional[Mapping[str, Any]]) -> Any:
        obj = self._convert_mapping(instance, '_from_validated_instance')
        return self.factory(**obj) if self.factory is not None and obj is not None else obj

    def _convert_mapping(self, mapping_or_obj: Optional[Mapping[str, Any]], method_name: str) \
            -> Optional[Mapping[str, Any]]:
        # TODO: recognise self.dependencies. if dependency is again a schema, compute effective schema
        #       by merging this and the dependency.

        if mapping_or_obj is None:
            return None

        # Fix for #432: call object's own to_dict() method but beware of infinite recursion
        if (hasattr(mapping_or_obj, 'to_dict')
                and callable(mapping_or_obj.to_dict)
                # calling JsonSchema.to_dict() is always fine
                and (isinstance(mapping_or_obj, JsonSchema)
                     # calling JsonObject.to_dict() would cause infinite recursion
                     or not isinstance(mapping_or_obj, JsonObject))):
            # noinspection PyBroadException
            try:
                return mapping_or_obj.to_dict()
            except BaseException:
                pass

        required_set = set(self.required) if self.required else set()

        additional_properties = self.additional_properties
        if additional_properties is None:
            # As of JSON Schema, additional_properties defaults to True
            additional_properties = True

        additional_properties_schema = None
        if isinstance(additional_properties, JsonSchema):
            additional_properties_schema = additional_properties

        converted_mapping = dict()

        if isinstance(mapping_or_obj, (collections.abc.Mapping, dict)):
            # An object that is already a Mapping
            mapping = mapping_or_obj
        else:
            # An object that is not a Mapping: convert to mapping.
            mapping = {}
            for property_name in dir(mapping_or_obj):
                property_value = UNDEFINED
                property_ok = False
                if property_name in self.properties:
                    # If property_name is defined in properties, we fully trust it
                    property_value = getattr(mapping_or_obj, property_name)
                    property_ok = True
                elif additional_properties and not property_name.startswith('_'):
                    # If property_name is not defined in properties, but additional
                    # properties are allowed, filter out private variables and callables.
                    property_value = getattr(mapping_or_obj, property_name)
                    property_ok = not callable(property_value)
                property_ok = property_ok and (property_name in required_set or property_value is not None)
                if property_ok and property_value != UNDEFINED:
                    mapping[property_name] = property_value

        # recursively convert mapping into converted_mapping according to schema

        # process defined properties
        for property_name, property_schema in self.properties.items():
            if property_name in mapping:
                property_value = mapping[property_name]
                converted_property_value = getattr(property_schema, method_name)(property_value)
                converted_mapping[property_name] = converted_property_value
            else:
                property_value = property_schema.default
                if property_value != UNDEFINED:
                    converted_property_value = getattr(property_schema, method_name)(property_value)
                    if property_name in required_set or converted_property_value is not None:
                        converted_mapping[property_name] = converted_property_value

        # process additional properties
        if additional_properties:
            for property_name, property_value in mapping.items():
                if property_name not in converted_mapping and property_value != UNDEFINED:
                    if additional_properties_schema:
                        property_value = getattr(additional_properties_schema, method_name)(property_value)
                    converted_mapping[property_name] = property_value

        return converted_mapping

    @classmethod
    def inject_attrs(cls, obj: object, attrs: Dict[str, Any]):
        for k, v in (attrs or {}).items():
            setattr(obj, k, v)

class JsonObject(ABC):
    """
    The abstract base class for objects

    * whose instances can be created from a JSON-serializable
      dictionary using their :meth:from_dict class method;
    * whose instances can be converted into a JSON-serializable dictionary
      using their :meth:to_dict instance method.

    Derived concrete classes must only implement the :meth:get_schema class method
    that must return a :class:JsonObjectSchema.

    Instances of this class have a JSON representation in Jupyter/IPython Notebooks.
    """

    @classmethod
    @abstractmethod
    def get_schema(cls) -> JsonObjectSchema:
        """Get JSON object schema."""

    @classmethod
    def from_dict(cls, value: Dict[str, Any]) -> 'JsonObject':
        """Create instance from JSON-serializable dictionary *value*."""
        return cls.get_schema().from_instance(value)

    def to_dict(self) -> Dict[str, Any]:
        """Create JSON-serializable dictionary representation."""
        return self.get_schema().to_instance(self)

    def _inject_attrs(self, attrs: Dict[str, Any]):
        assert_instance(attrs, dict, name='attrs')
        schema = self.get_schema()
        assert_true(isinstance(schema, JsonObjectSchema),
                    message='schema must be a JSON object schema')
        all_attrs = {k: None for k in (schema.properties or {}).keys()}
        all_attrs.update(attrs)
        JsonObjectSchema.inject_attrs(self, all_attrs)


register_json_formatter(JsonSchema)
register_json_formatter(JsonObject)



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import Optional, Mapping, Dict

from xcube.constants import LOG
from xcube.core.tile import DEFAULT_CRS_NAME
from xcube.core.tile import DEFAULT_FORMAT
from xcube.core.tile import TileNotFoundException
from xcube.core.tile import TileRequestException
from xcube.core.tile import compute_rgba_tile
from xcube.core.tilingscheme import DEFAULT_TILE_SIZE
from xcube.server.api import ApiError
from xcube.util.perf import measure_time_cm
from .context import TilesContext


def compute_ml_dataset_tile(ctx: TilesContext,
                            ds_id: str,
                            var_name: str,
                            crs_name: Optional[str],
                            x: str, y: str, z: str,
                            params: Mapping[str, str]):
    params = dict(params)
    trace_perf = params.pop(
        'debug',
        '1' if ctx.datasets_ctx.trace_perf else '0'
    ) == '1'
    measure_time = measure_time_cm(logger=LOG, disabled=not trace_perf)
    with measure_time('Computing RGBA tile'):
        return _compute_ml_dataset_tile(ctx,
                                        ds_id,
                                        var_name,
                                        crs_name,
                                        x, y, z,
                                        params,
                                        trace_perf)


def _compute_ml_dataset_tile(ctx: TilesContext,
                             ds_id: str,
                             var_name: str,
                             crs_name: Optional[str],
                             x: str, y: str, z: str,
                             args: Dict[str, str],
                             trace_perf: bool):
    try:
        x, y, z = [int(c) for c in (x, y, z)]
    except ValueError:
        raise ApiError.BadRequest('x, y, z must be integers')

    crs_name = args.pop('crs', crs_name or DEFAULT_CRS_NAME)
    retina = args.pop('retina', None) == '1'
    cmap_name = args.pop('cmap', args.pop('cbar', None))
    value_min = float(args.pop('vmin', 0.0))
    value_max = float(args.pop('vmax', 1.0))
    format = args.pop('format', DEFAULT_FORMAT)

    if format not in ('png', 'image/png'):
        raise ApiError.BadRequest(
            f'Illegal format {format!r}'
        )

    ml_dataset = ctx.datasets_ctx.get_ml_dataset(ds_id)
    if var_name == 'rgb':
        var_names, value_ranges = ctx.datasets_ctx.get_rgb_color_mapping(
            ds_id, norm_range=(value_min, value_max)
        )
        components = 'r', 'g', 'b'
        for i, c in enumerate(components):
            var_names[i] = args.pop(c, var_names[i])
            if var_names[i] is None:
                raise ApiError.BadRequest(
                    f'No variable in dataset {ds_id!r}'
                    f' specified for RGB component {c.upper()}'
                )
            value_ranges[i] = (
                float(args.pop(
                    f'{c}vmin', value_ranges[i][0]
                )),
                float(args.pop(
                    f'{c}vmax', value_ranges[i][1]
                ))
            )
    else:
        if cmap_name is None or value_min is None or value_max is None:
            default_cmap_name, (default_value_min, default_value_min) = \
                ctx.datasets_ctx.get_color_mapping(ds_id, var_name)
            if cmap_name is None:
                cmap_name = default_cmap_name
            if value_min is None:
                value_min = default_value_min
            if value_max is None:
                value_max = default_value_min
        var_names = (var_name,)
        value_ranges = ((value_min, value_max),)

    try:
        return compute_rgba_tile(
            ml_dataset,
            var_names,
            x, y, z,
            ctx.datasets_ctx.colormap_registry,
            crs_name=crs_name,
            tile_size=(2 if retina else 1) * DEFAULT_TILE_SIZE,
            cmap_name=cmap_name,
            value_ranges=value_ranges,
            non_spatial_labels=args,
            format=format,
            trace_perf=trace_perf
        )
    except TileNotFoundException as e:
        raise ApiError.NotFound(f'{e}') from e
    except TileRequestException as e:
        raise ApiError.BadRequest(f'{e}') from e



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import pkgutil
import sys
from string import Template

from xcube.server.api import ApiError
from xcube.server.api import ApiHandler
from .api import api
from .context import MetaContext
from .controllers import get_service_info
from ...constants import LOG


@api.route("/")
class ServiceInfoHandler(ApiHandler[MetaContext]):
    @api.operation(operation_id='getServiceInfo',
                   summary='Get information about the service')
    def get(self):
        self.response.finish(get_service_info(self.ctx))


@api.route("/openapi.html")
class OpenApiHtmlHandler(ApiHandler):
    @api.operation(
        operation_id='getOpenApiHtml',
        summary='Show API documentation'
    )
    def get(self):
        include_all = self.request.get_query_arg('all', default=False)
        html_template = pkgutil.get_data('xcube.webapi.meta',
                                         'data/openapi.html').decode('utf-8')
        self.response.finish(Template(html_template).substitute(
            open_api_url=self.request.url_for_path(
                'openapi.json',
                query='all=1' if include_all else None,
                reverse=True
            )
        ))


@api.route("/openapi.json")
class OpenApiJsonHandler(ApiHandler):
    @api.operation(
        operation_id='getOpenApiJson',
        summary='Get API documentation as OpenAPI 3.0 JSON document'
    )
    def get(self):
        include_all = self.request.get_query_arg('all', default=False)
        self.response.finish(
            self.ctx.get_open_api_doc(include_all=include_all),
            content_type='application/vnd.oai.openapi+json;version=3.0'
        )


@api.route("/maintenance/fail")
class MaintenanceFailHandler(ApiHandler[MetaContext]):
    @api.operation(operation_id='forceError',
                   summary='Force a request error (for testing)',
                   parameters=[
                       {
                           "name": "code",
                           "in": "query",
                           "description": "HTTP status code",
                           "schema": {
                               "type": "integer",
                               "minimum": 400,
                           }
                       },
                       {
                           "name": "message",
                           "in": "query",
                           "description": "HTTP error message",
                           "schema": {
                               "type": "string",
                           }
                       },
                   ])
    def get(self):
        """If *code* is given, the operation fails with
        that HTTP status code. Otherwise, the operation causes
        an internal server error.
        """
        code = self.request.get_query_arg('code', type=int)
        message = self.request.get_query_arg(
            'message',
            default='Error! No worries, this is just a test.'
        )
        LOG.warning("Forcing error in request...")
        if code is None:
            raise RuntimeError(message)
        else:
            raise ApiError(code, message=message)


@api.route("/maintenance/update")
class MaintenanceUpdateHandler(ApiHandler[MetaContext]):
    @api.operation(operation_id='updateServer',
                   summary='Force server update,'
                           ' updates configuration, and'
                           ' resets all resource caches.')
    async def get(self):
        LOG.warning("Forcing server update...")
        # self.ctx.call_later(0,
        #                     self.ctx.server_ctx.server.update,
        #                     self.ctx.config)
        # self.ctx.server_ctx.server.update(self.ctx.config)
        await self.ctx.run_in_executor(None,
                                       self.ctx.server_ctx.server.update,
                                       self.ctx.config)
        return self.response.finish(dict(status="OK"))


@api.route("/maintenance/kill")
class MaintenanceUpdateHandler(ApiHandler[MetaContext]):
    @api.operation(operation_id='killServer',
                   summary='Force server to shut down immediately.')
    def get(self):
        LOG.warning("Forcing server death...")
        sys.exit(0)


@api.route("/maintenance/oom")
class MaintenanceUpdateHandler(ApiHandler[MetaContext]):
    @api.operation(operation_id='forceOutOfMemory',
                   summary='Force an out-of-memory error (for testing).')
    def get(self):
        LOG.warning("Forcing out-of-memory error...")
        my_bytes = []
        from itertools import count
        for i in count(start=0, step=1):
            my_bytes.append(bytes(2 ** i))



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.


from typing import Dict, Union

import pyproj

CrsLike = Union[str, pyproj.CRS]


class ProjCache:
    """
    A cache for pyproj objects
    that may take considerable time to construct.
    """
    INSTANCE: 'ProjCache'

    def __init__(self):
        self._crs_cache: Dict[str, pyproj.CRS] = dict()
        self._transformer_cache: Dict[str, pyproj.Transformer] = dict()

    def get_crs(self, crs: CrsLike) -> pyproj.CRS:
        if isinstance(crs, pyproj.CRS):
            return crs
        key = crs
        if key not in self._crs_cache:
            # pyproj.CRS.from_string() is expensive
            # save result for later use
            self._crs_cache[key] = pyproj.CRS.from_string(crs)
        return self._crs_cache[key]

    def get_transformer(self,
                        crs1: CrsLike,
                        crs2: CrsLike) -> pyproj.Transformer:
        crs1_key = self.get_crs_srs(crs1)
        crs2_key = self.get_crs_srs(crs2)
        key = f'{crs1_key}->{crs2_key}'
        if key not in self._transformer_cache:
            crs1 = self.get_crs(crs1)
            crs2 = self.get_crs(crs2)
            # pyproj.Transformer.from_crs() is really expensive (~0.5 secs)
            # save result for later use
            self._transformer_cache[key] = pyproj.Transformer.from_crs(
                crs1, crs2, always_xy=True)
        return self._transformer_cache[key]

    @classmethod
    def get_crs_srs(cls, crs: CrsLike) -> str:
        if isinstance(crs, pyproj.CRS):
            return crs.srs
        return crs


ProjCache.INSTANCE = ProjCache()



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from xcube.server.api import ApiHandler, ApiError

from .api import api
from .context import StacContext
from .controllers import get_collection, get_collection_queryables, \
    get_single_collection_items, \
    get_collection_schema
from .controllers import get_collection_item
from .controllers import get_datasets_collection_items
from .controllers import get_collections
from .controllers import get_conformance
from .controllers import get_root
from .controllers import search
from .config import (
    PATH_PREFIX,
    DEFAULT_COLLECTION_ID,
)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX)
@api.route(PATH_PREFIX + "/")
class CatalogRootHandler(ApiHandler[StacContext]):
    @api.operation(operation_id="getCatalogRoot",
                   summary="Get the STAC catalog's root.")
    async def get(self):
        result = get_root(self.ctx.datasets_ctx,
                          self.request.reverse_base_url)
        return await self.response.finish(result)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/conformance")
@api.route(PATH_PREFIX + "/conformance/")
class CatalogConformanceHandler(ApiHandler[StacContext]):
    @api.operation(operation_id="getCatalogConformance",
                   summary="Get the STAC catalog's conformance.")
    async def get(self):
        result = get_conformance()
        return await self.response.finish(result)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/collections")
@api.route(PATH_PREFIX + "/collections/")
class CatalogCollectionsHandler(ApiHandler[StacContext]):
    @api.operation(operation_id="getCatalogCollections",
                   summary="Get the STAC catalog's collections.")
    async def get(self):
        result = get_collections(self.ctx, self.request.reverse_base_url)
        return await self.response.finish(result)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/collections/{collectionId}")
@api.route(PATH_PREFIX + "/collections/{collectionId}/")
class CatalogCollectionHandler(ApiHandler[StacContext]):
    # noinspection PyPep8Naming
    @api.operation(operation_id="getCatalogCollection",
                   summary="Get a STAC catalog collection.")
    async def get(self, collectionId: str):
        result = get_collection(self.ctx,
                                self.request.reverse_base_url,
                                collectionId)
        return await self.response.finish(result)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/collections/{collectionId}/items")
@api.route(PATH_PREFIX + "/collections/{collectionId}/items/")
class CatalogCollectionItemsHandler(ApiHandler[StacContext]):
    # noinspection PyPep8Naming
    @api.operation(operation_id="getCatalogCollectionItems",
                   summary="Get the items of a STAC catalog collection.")
    async def get(self, collectionId: str):
        get_items_args = dict(
            ctx=self.ctx.datasets_ctx,
            base_url=self.request.reverse_base_url,
            collection_id=collectionId,
        )
        if collectionId == DEFAULT_COLLECTION_ID:
            if 'limit' in self.request.query:
                get_items_args['limit'] = \
                    self.request.get_query_arg('limit', type=int)
            if 'cursor' in self.request.query:
                get_items_args['cursor'] = \
                    self.request.get_query_arg('cursor', type=int)
            result = get_datasets_collection_items(**get_items_args)
        else:
            result = get_single_collection_items(**get_items_args)

        return await self.response.finish(
            result, content_type='application/geo+json'
        )


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + '/collections/{collectionId}/items/{featureId}')
@api.route(PATH_PREFIX + '/collections/{collectionId}/items/{featureId}/')
class CatalogCollectionItemHandler(ApiHandler[StacContext]):
    # noinspection PyPep8Naming
    @api.operation(
        operation_id="getCatalogCollectionItem",
        summary="Get an item of a STAC catalog collection.",
    )
    async def get(self, collectionId: str, featureId: str):
        result = get_collection_item(
            self.ctx.datasets_ctx,
            self.request.reverse_base_url,
            collectionId,
            featureId,
        )
        return await self.response.finish(
            result, content_type='application/geo+json'
        )


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/search")
@api.route(PATH_PREFIX + "/search/")
class CatalogSearchHandler(ApiHandler[StacContext]):

    @api.operation(operation_id="searchCatalogByKeywords",
                   summary="Search the STAC catalog by keywords.")
    async def get(self):
        # TODO (forman): get search params from query
        result = search(self.ctx.datasets_ctx, self.request.reverse_base_url)
        return await self.response.finish(result)

    @api.operation(operation_id="searchCatalogByJSON",
                   summary="Search the STAC catalog by JSON request body.")
    async def post(self):
        # TODO (forman): get search params from request body
        result = search(self.ctx.datasets_ctx, self.request.reverse_base_url)
        return await self.response.finish(result)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/collections/{collectionId}/queryables")
@api.route(PATH_PREFIX + "/collections/{collectionId}/queryables/")
class QueryablesHandler(ApiHandler[StacContext]):
    # noinspection PyPep8Naming
    @api.operation(operation_id='queryables',
                   summary='Return a JSON Schema defining the supported '
                           'metadata filters (also called "queryables") '
                           'for a specific collection.')
    async def get(self, collectionId):
        schema = get_collection_queryables(
            ctx=self.ctx.datasets_ctx, collection_id=collectionId
        )
        return await self.response.finish(schema)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/collections/{collectionId}/schema")
@api.route(PATH_PREFIX + "/collections/{collectionId}/schema/")
class SchemaHandler(ApiHandler[StacContext]):
    # noinspection PyPep8Naming
    @api.operation(operation_id='schema',
                   summary='Return a JSON Schema defining the range '
                           '(i.e. data variables) of a specific collection.')
    async def get(self, collectionId):
        schema = get_collection_schema(
            ctx=self.ctx.datasets_ctx,
            base_url=self.request.reverse_base_url,
            collection_id=collectionId
        )
        return await self.response.finish(schema)



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import os.path
from functools import cache
from typing import Mapping, Any, Optional

from xcube.constants import DEFAULT_SERVER_ADDRESS
from xcube.constants import DEFAULT_SERVER_PORT
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonBooleanSchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema

BASE_SERVER_CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        port=JsonIntegerSchema(
            title='Server port.',
            default=DEFAULT_SERVER_PORT
        ),
        address=JsonStringSchema(
            title='Server address.',
            default=DEFAULT_SERVER_ADDRESS
        ),
        base_dir=JsonStringSchema(
            title='Base directory used to resolve relative local paths.'
                  ' Can be a local filesystem path or an absolute URL.',
        ),
        url_prefix=JsonStringSchema(
            title='Prefix to be prepended to all URL route paths.'
                  ' Can be an absolute URL or a relative URL path.',
        ),
        reverse_url_prefix=JsonStringSchema(
            title='Prefix to be prepended to reverse URL paths'
                  ' returned by server responses.'
                  ' Can be an absolute URL or a relative URL path.',
        ),
        trace_perf=JsonBooleanSchema(
            title='Output performance measures',
        ),
        static_routes=JsonArraySchema(
            title='Static content routes',
            items=JsonObjectSchema(
                properties=dict(
                    path=JsonStringSchema(
                        title='The URL path',
                        min_length=1
                    ),
                    dir_path=JsonStringSchema(
                        title='A local directory path',
                        min_length=1
                    ),
                    default_filename=JsonStringSchema(
                        title='Optional default filename',
                        examples=['index.html'],
                        min_length=1
                    ),
                    openapi_metadata=JsonObjectSchema(
                        title='Optional OpenAPI operation metadata',
                        additional_properties=True
                    ),
                ),
                required=['path', 'dir_path'],
                additional_properties=False
            )
        ),
        api_spec=JsonObjectSchema(
            title='API specification',
            description='selected = (includes | ALL) - (excludes | NONE)',
            properties=dict(
                includes=JsonArraySchema(
                    JsonStringSchema(min_length=1)
                ),
                excludes=JsonArraySchema(
                    JsonStringSchema(min_length=1)
                ),
            ),
            additional_properties=False
        )
    ),
    # We allow for other configuration settings contributed
    # by APIs. If these APIs are currently not in use,
    # validation would fail if additional_properties=False.
    additional_properties=True,
)


@cache
def normalize_base_dir(base_dir: Optional[str]) -> str:
    """Normalize the given base directory *base_dir*.
    """
    if base_dir is None:
        base_dir = os.path.abspath("")
    elif not is_absolute_path(base_dir):
        base_dir = os.path.abspath(base_dir)
    while base_dir != '/' and base_dir.endswith('/'):
        base_dir = base_dir[:-1]
    return base_dir


def get_base_dir(config: Mapping[str, Any]) -> str:
    """Get the normalized base directory from configuration *config*.
    """
    return normalize_base_dir(config.get("base_dir"))


def resolve_config_path(config: Mapping[str, Any], path: str) -> str:
    """Resolve a given relative *path* against the base directory given by
    *config*. Return *path* unchanged, if it is absolute.
    """
    if is_absolute_path(path):
        abs_path = path
    else:
        base_dir = get_base_dir(config)
        abs_path = f"{base_dir}/{path}"
    # Resolve ".." and "." in path
    if "://" in abs_path:
        scheme, host_path = abs_path.split("://", maxsplit=1)
        if "/" in host_path:
            hostname, url_path = host_path.split("/", maxsplit=1)
            url_path = _remove_path_dot_segments(url_path)
            return f"{scheme}://{hostname}/{url_path}"
        else:
            return f"{scheme}://{host_path}"
    else:
        if os.name == "nt":
            # Windows can also live with forward slashes
            abs_path = abs_path.replace("\\", "/")
        return _remove_path_dot_segments(abs_path)


#
# Following code is stolen from urllib3.url
#
def _remove_path_dot_segments(path: str) -> str:
    # See http://tools.ietf.org/html/rfc3986#section-5.2.4 for pseudo-code
    segments = path.split("/")  # Turn the path into a list of segments
    output = []  # Initialize the variable to use to store output

    for segment in segments:
        # '.' is the current directory, so ignore it, it is superfluous
        if segment == ".":
            continue
        # Anything other than '..', should be appended to the output
        elif segment != "..":
            output.append(segment)
        # In this case segment == '..', if we can, we should pop the last
        # element
        elif output:
            output.pop()

    # If the path starts with '/' and the output is empty or the first string
    # is non-empty
    if path.startswith("/") and (not output or output[0]):
        output.insert(0, "")

    # If the path starts with '/.' or '/..' ensure we add one more empty
    # string to add a trailing '/'
    if path.endswith(("/.", "/..")):
        output.append("")

    return "/".join(output)


def is_absolute_path(path: str) -> bool:
    """Test whether *path* is an absolute filesystem path or URL.
    """
    # This is a rather weak test, may be enhanced if desired
    return "//" in path \
           or ":" in path \
           or path.startswith("/")


def get_url_prefix(config: Mapping[str, Any]) -> str:
    """Get the sanitized URL prefix so, if given, it starts with
    a leading slash and ends without one.

    :param config: Server configuration.
    :return: Sanitized URL prefix, may be an empty string.
    """
    return _sanitize_url_prefix(config.get('url_prefix'))


def get_reverse_url_prefix(config: Mapping[str, Any]) -> str:
    """Get the sanitized reverse URL prefix so, if given, it starts with
    a leading slash and ends without one.

    :param config: Server configuration.
    :return: Sanitized URL prefix, may be an empty string.
    """
    return _sanitize_url_prefix(
        config.get('reverse_url_prefix',
                   config.get('url_prefix'))
    )


def _sanitize_url_prefix(url_prefix: Optional[str]) -> str:
    """Get a sanitized URL prefix so, if given, it starts with
    a leading slash and ends without one.

    :param url_prefix: URL prefix path.
    :return: Sanitized URL prefix path, may be an empty string.
    """
    if not url_prefix:
        return ''

    while url_prefix.startswith('//'):
        url_prefix = url_prefix[1:]
    while url_prefix.endswith('/'):
        url_prefix = url_prefix[:-1]

    if url_prefix == '':
        return ''

    if url_prefix.startswith('/') \
            or url_prefix.startswith('http://') \
            or url_prefix.startswith('https://'):
        return url_prefix

    return '/' + url_prefix
